import logging
import pathlib
from collections import defaultdict
from collections.abc import Generator
from unittest.mock import Mock, patch
from zipfile import ZipFile

import pytest
from natsort import natsorted
from neuxml import xmlmap

from remarx.sentence.corpus.alto_input import (
    AltoDocument,
    ALTOInput,
    TextBlock,
    TextLine,
)
from remarx.sentence.corpus.base_input import FileInput, SectionType
from test_sentence.test_corpus.test_text_input import simple_segmenter

FIXTURE_DIR = pathlib.Path(__file__).parent / "fixtures"
FIXTURE_ALTO_ZIPFILE = FIXTURE_DIR / "alto_sample.zip"
FIXTURE_ALTO_PAGE = FIXTURE_DIR / "alto_page.xml"

# test xmlmap classes


def test_alto_document():
    altoxml = xmlmap.load_xmlobject_from_file(FIXTURE_ALTO_PAGE, AltoDocument)
    # sample page has 4 text blocks
    assert len(altoxml.blocks) == 4
    assert isinstance(altoxml.blocks[0], TextBlock)


def test_alto_document_is_alto():
    # sample page is alto
    altoxml = xmlmap.load_xmlobject_from_file(FIXTURE_ALTO_PAGE, AltoDocument)
    assert altoxml.is_alto()

    # load tei fixture as alto document to check non-alto content
    teixml = xmlmap.load_xmlobject_from_file(
        FIXTURE_DIR / "sample_tei.xml", AltoDocument
    )
    assert not teixml.is_alto()


def test_alto_document_sorted_blocks():
    altoxml = xmlmap.load_xmlobject_from_file(FIXTURE_ALTO_PAGE, AltoDocument)
    # sample alto page has been rearranged so text blocks are not sequential
    # by vertical position
    assert altoxml.blocks != altoxml.sorted_blocks
    # first block was moved to second, so 1 unsorted should match 0 sorted
    assert altoxml.blocks[1] == altoxml.sorted_blocks[0]
    # first block should be the smallest vpos on the page
    min_vpos = min(tb.vertical_position for tb in altoxml.blocks)
    assert altoxml.sorted_blocks[0].vertical_position == min_vpos

    # sorted logic when no text blocks
    empty_alto = xmlmap.load_xmlobject_from_string("<root/>", AltoDocument)
    assert empty_alto.sorted_blocks == []


def test_alto_document_text_chunks():
    altoxml = xmlmap.load_xmlobject_from_file(FIXTURE_ALTO_PAGE, AltoDocument)
    chunks = altoxml.text_chunks()
    assert isinstance(chunks, Generator)
    # convert to list to verify contents
    chunks = list(chunks)
    # should be list of dict
    assert isinstance(chunks[0], dict)
    # should be one dict per text block
    assert len(chunks) == len(altoxml.blocks)
    assert chunks[0]["text"] == altoxml.sorted_blocks[0].text_content
    # for now, everything is text
    assert chunks[0]["section_type"] == SectionType.TEXT.value


def test_alto_textblock():
    altoxml = xmlmap.load_xmlobject_from_file(FIXTURE_ALTO_PAGE, AltoDocument)
    alto_textblock = altoxml.blocks[1]
    assert alto_textblock.horizontal_position == 728.0
    assert alto_textblock.vertical_position == 200.0
    assert len(alto_textblock.lines) == 1
    assert isinstance(alto_textblock.lines[0], TextLine)


def test_alto_textblock_sorted_lines():
    altoxml = xmlmap.load_xmlobject_from_file(FIXTURE_ALTO_PAGE, AltoDocument)
    # the third text block has the most lines;
    # lines 1 & 2 manually moved to force out of order
    alto_textblock = altoxml.blocks[2]
    assert alto_textblock.lines != alto_textblock.sorted_lines
    # first line  was moved to second, so 1 unsorted should match 0 sorted
    assert alto_textblock.lines[1] == alto_textblock.sorted_lines[0]
    # first line should be the smallest vpos on the page
    min_vpos = min(line.vertical_position for line in alto_textblock.lines)
    assert alto_textblock.sorted_lines[0].vertical_position == min_vpos


def test_alto_textblock_text_content():
    altoxml = xmlmap.load_xmlobject_from_file(FIXTURE_ALTO_PAGE, AltoDocument)

    alto_textblock = altoxml.blocks[3]
    block_text = alto_textblock.text_content
    # check that we have the expected number of lines
    assert len(block_text.split("\n")) == len(alto_textblock.lines)
    # and content starts and ends with expected contents
    assert block_text.startswith("Ist gegen die Klausel")  # codespell:ignore
    assert block_text.endswith("1896-97. I. Bd.")


def test_alto_textline():
    altoxml = xmlmap.load_xmlobject_from_file(FIXTURE_ALTO_PAGE, AltoDocument)
    alto_textline = altoxml.blocks[1].lines[0]
    assert alto_textline.horizontal_position == 868.0
    assert alto_textline.vertical_position == 256.0
    assert (
        alto_textline.text_content
        == "F. A. Sorge: Die Präsidentenwahl in den Vereinigten Staaten."
    )
    assert str(alto_textline) == alto_textline.text_content


# test file input classes


def test_field_names():
    assert ALTOInput.field_names == (*FileInput.field_names, "section_type")


def test_altoinput_get_text(caplog):
    caplog.set_level(logging.INFO, logger="remarx.sentence.corpus.alto_input")
    alto_input = ALTOInput(input_file=FIXTURE_ALTO_ZIPFILE)
    chunks = alto_input.get_text()

    # confirm generator type, then convert to list to inspect results
    assert isinstance(chunks, Generator)
    chunks = list(chunks)
    assert isinstance(chunks[0], dict)

    expected_files = [
        "1896-97a.pdf_page_1.xml",
        "1896-97a.pdf_page_2.xml",
        "1896-97a.pdf_page_3.xml",
        "1896-97a.pdf_page_4.xml",
        "1896-97a.pdf_page_5.xml",
        "empty_page.xml",
        "unsorted_page.xml",
    ]

    # distinct filenames should match expected file list
    chunks_by_filename = defaultdict(list)
    for chunk in chunks:
        chunks_by_filename[chunk["file"]].append(chunk)

    assert set(chunks_by_filename.keys()) == set(expected_files)
    # all sections are currently text
    assert {chunk["section_type"] for chunk in chunks} == {SectionType.TEXT.value}

    # inspect text results for a few cases
    assert (
        chunks_by_filename["1896-97a.pdf_page_3.xml"][0]["text"]
        == "Arbeiter und Gewerbeausstellung."
    )
    assert chunks_by_filename["1896-97a.pdf_page_3.xml"][2]["text"].endswith(
        "langsamer und deshalb auch viel häßlicher und viel widerlicher. Und wie die"
    )

    processing_prefix = "Processing XML file "
    processed_files = [
        record.getMessage().removeprefix(processing_prefix)
        for record in caplog.records
        if record.getMessage().startswith(processing_prefix)
    ]
    assert processed_files == natsorted(expected_files)

    # last log entry should report time to process, # of files
    summary_log_message = caplog.records[-1].getMessage()
    assert summary_log_message.startswith(
        f"Processed {FIXTURE_ALTO_ZIPFILE.name} with 7 files (7 valid ALTO)"
    )


def test_altoinput_warn_no_text(caplog):
    alto_input = ALTOInput(input_file=FIXTURE_ALTO_ZIPFILE)
    with caplog.at_level(logging.WARNING, logger="remarx.sentence.corpus.alto_input"):
        list(alto_input.get_text())

    warning_messages = [record.getMessage() for record in caplog.records]
    assert any(
        message == "No text lines found in ALTO XML file: empty_page.xml"
        for message in warning_messages
    )


def test_altoinput_error_non_xml(tmp_path: pathlib.Path):
    archive_path = tmp_path / "invalid.zip"
    with ZipFile(archive_path, "w") as archive:
        archive.writestr("page1.txt", "not xml file")

    alto_input = ALTOInput(input_file=archive_path)
    with pytest.raises(
        ValueError, match=f"No valid ALTO XML files found in {archive_path.name}"
    ):
        list(alto_input.get_text())


def test_altoinput_error_non_alto_xml(tmp_path: pathlib.Path):
    archive_path = tmp_path / "not_alto.zip"
    with ZipFile(archive_path, "w") as archive:
        archive.writestr("page1.xml", "<root></root>")

    alto_input = ALTOInput(input_file=archive_path)
    with pytest.raises(
        ValueError, match=f"No valid ALTO XML files found in {archive_path.name}"
    ):
        list(alto_input.get_text())


def test_altoinput_error_non_alto_xml_unknown_namespace(tmp_path: pathlib.Path):
    archive_path = tmp_path / "unknown_ns.zip"
    xml_content = '<alto xmlns="http://unknown_namespace.com/alto/ns#"><Description></Description></alto>'
    with ZipFile(archive_path, "w") as archive:
        archive.writestr("page1.xml", xml_content)

    alto_input = ALTOInput(input_file=archive_path)
    with pytest.raises(
        ValueError, match=f"No valid ALTO XML files found in {archive_path.name}"
    ):
        list(alto_input.get_text())


def test_altoinput_warn_invalid_xml(tmp_path: pathlib.Path, caplog):
    archive_path = tmp_path / "invalid_xml.zip"
    with ZipFile(archive_path, "w") as archive:
        archive.writestr("page1.xml", "<alto>")

    alto_input = ALTOInput(input_file=archive_path)
    caplog.set_level(logging.DEBUG, logger="remarx.sentence.corpus.alto_input")
    with pytest.raises(
        ValueError, match=f"No valid ALTO XML files found in {archive_path.name}"
    ):
        list(alto_input.get_text())

    # expect one warning message (skipping the file) and one debug (xml syntax error)
    warning_message = next(
        rec.getMessage() for rec in caplog.records if rec.levelname == "WARNING"
    )
    debug_message = next(
        rec.getMessage() for rec in caplog.records if rec.levelname == "DEBUG"
    )
    assert warning_message == "Skipping page1.xml : invalid XML"
    # debug message includes info about the syntax error
    assert "XML syntax error" in debug_message
    assert "Premature end of data in tag alto line 1" in debug_message


def test_altoinput_error_empty_zip(tmp_path: pathlib.Path):
    archive_path = tmp_path / "empty.zip"
    # create an empty but valid zipfile
    with ZipFile(archive_path, "w"):
        pass

    alto_input = ALTOInput(input_file=archive_path)
    with pytest.raises(
        ValueError, match=f"No valid ALTO XML files found in {archive_path.name}"
    ):
        list(alto_input.get_text())


@patch("remarx.sentence.corpus.base_input.segment_text")
def test_get_sentences_sequntial(mock_segment_text: Mock):
    # patch in simple segmenter to split each input text in two
    mock_segment_text.side_effect = simple_segmenter

    alto_input = ALTOInput(input_file=FIXTURE_ALTO_ZIPFILE)
    sentences = list(alto_input.get_sentences())
    num_sentences = len(sentences)
    # currently with this fixture data and simple segmenter,  expect 40 sentences
    assert num_sentences == 40

    # sentence indexes should start at 0 and continue across all sentences
    indexes = [sentence["sent_index"] for sentence in sentences]
    assert indexes == list(range(num_sentences))



"Input class for handling basic text file as input for corpus creation."

from collections.abc import Generator
from dataclasses import dataclass

from remarx.sentence.corpus.base_input import FileInput


@dataclass
class TextInput(FileInput):
    """
    Basic text file input handling for sentence corpus creation. Takes
    a single text input file and returns text without chunking.
    """

    file_type = ".txt"
    "Supported file extension for text input"

    def get_text(self) -> Generator[dict[str, str]]:
        """
        Get plain-text contents for this file with any desired chunking (e.g.
        pages or other semantic unit).
        Default implementation does no chunking, no additional metadata.

        :returns: Generator with a dictionary of text and any other metadata
        that applies to this unit of text.
        """
        yield {"text": self.input_file.read_text(encoding="utf-8")}



"""
Library for finding sentence-level quote pairs.

Note: Currently this script only supports one original and reuse corpus.
"""

import logging
import pathlib
from timeit import default_timer as time

import numpy.typing as npt
import polars as pl
from voyager import Index, Space

from remarx.quotation.embeddings import get_sentence_embeddings

logger = logging.getLogger(__name__)


def build_vector_index(embeddings: npt.NDArray) -> Index:
    """
    Builds an index for a given set of embeddings with the specified
    number of trees.
    """
    # TODO: Add relevant voyager parameters
    start = time()
    # Instantiate annoy index using dot product
    n_vecs, n_dims = embeddings.shape
    index = Index(Space.InnerProduct, num_dimensions=n_dims, max_elements=n_vecs)
    # more efficient to add all vectors at once
    index.add_items(embeddings)
    # Return the index
    # NOTE: index could be saved to disk, which may be helpful in future
    elapsed_time = time() - start
    logger.info(
        f"Created index with {index.num_elements} items and {n_dims} dimensions in {elapsed_time:.1f} seconds"
    )
    return index


def get_sentence_pairs(
    original_sents: list[str],
    reuse_sents: list[str],
    score_cutoff: float,
    show_progress_bar: bool = False,
) -> pl.DataFrame:
    """
    For a set of original and reuse sentences, identify pairs of original-reuse
    sentence pairs where quotation is likely. Returns these sentence pairs as
    a polars DataFrame including for each pair:

    - `original_index`: the index of the original sentence
    - `reuse_index`: the index of the reuse sentence
    - `match_score`: the quality of the match

    Likely quote pairs are identified through the sentences' embeddings. The Annoy
    library is used to find the nearest original sentence for each reuse sentence.
    Then likely quote pairs are determined by those sentence pairs with a match score
    (cosine similarity) above the specified cutoff.
    Optionally, the parameters for Annoy may be specified.
    """
    # Generate embeddings
    logger.info("Now generating sentence embeddings")
    start = time()
    original_vecs = get_sentence_embeddings(
        original_sents, show_progress_bar=show_progress_bar
    )
    reuse_vecs = get_sentence_embeddings(
        reuse_sents, show_progress_bar=show_progress_bar
    )
    n_vecs = len(original_vecs) + len(reuse_vecs)
    elapsed_time = time() - start
    logger.info(
        f"Generated {n_vecs:,} sentence embeddings in {elapsed_time:.1f} seconds"
    )

    # Build search index
    # NOTE: An index only needs to be generated once for a set of embeddings.
    #       Perhaps there's some potential reuse between runs?
    index = build_vector_index(original_vecs)

    # Get sentence matches; query all vectors at once
    # returns a list of lists with results for each reuse vector
    start = time()
    all_neighbor_ids, all_distances = index.query(reuse_vecs, k=1)
    elapsed_time = time() - start
    logger.info(
        f"Queried {len(reuse_vecs):,} sentence embeddings in {elapsed_time:.1f} seconds"
    )

    result = (
        pl.DataFrame(
            data={"original_index": all_neighbor_ids, "match_score": all_distances}
        )
        # add row index
        .with_row_index(name="reuse_index")
        # since we requested k=1, explode the lists to get single value result
        .explode("original_index", "match_score")
        # then filter by specified match score cutoff
        .filter(pl.col("match_score").lt(score_cutoff))
    )
    total = result.height
    logger.info(
        f"Identified {total:,} sentence pair{'' if total == 1 else 's'} under score cutuff {score_cutoff}"
    )
    return result


def load_sent_df(sentence_corpus: pathlib.Path, col_pfx: str = "") -> pl.DataFrame:
    """
    For a given sentence corpus, create a polars DataFrame suitable for finding
    sentence-level quote pairs. Optionally, a prefix can be added to all column names.

    The resulting dataframe has the same fields as the input corpus except with:

    - a new field `index` corresponding to the row index
    - the sentence id field `sent_id` is renamed to `id`

    """
    start_cols = ["index", "sent_id", "text"]
    return (
        # Since all required fields are strings, there's no need to infer schema
        pl.read_csv(sentence_corpus, row_index_name="index", infer_schema=False)
        .select(*start_cols, pl.all().exclude(start_cols))
        .rename({"sent_id": "id"})
        .rename(lambda x: f"{col_pfx}{x}")
    )


def compile_quote_pairs(
    original_corpus: pl.DataFrame,
    reuse_corpus: pl.DataFrame,
    detected_pairs: pl.DataFrame,
) -> pl.DataFrame:
    """
    Link sentence metadata to the detected sentence pairs from the given original
    and reuse sentence corpus dataframes to form quote pairs. The original and reuse
    corpus dataframes must contain a row index column named `original_index` and
    `reuse_index` respectively. Ideally, these dataframes should be built using
    [load_sent_df][remarx.quotation.pairs.load_sent_df].

    Returns a dataframe with the following fields:

    - `match_score`: Estimated quality of the match
    - All other fields in order from the reuse corpus except its row index
    - All other fields in order from the original corpus except its row index
    """
    # Build and return quote pairs
    return (
        detected_pairs.join(reuse_corpus, on="reuse_index")
        .join(original_corpus, on="original_index")
        .drop(["reuse_index", "original_index"])
    )


def find_quote_pairs(
    original_corpus: pathlib.Path,
    reuse_corpus: pathlib.Path,
    out_csv: pathlib.Path,
    score_cutoff: float = 0.225,
    show_progress_bar: bool = False,
) -> None:
    """
    For a given original and reuse sentence corpus, finds the likely sentence-level
    quote pairs. These quote pairs are saved as a CSV. Optionally, the required
    quality for quote pairs can be modified via `score_cutoff`.
    """

    # Build sentence dataframes
    original_df = load_sent_df(original_corpus, col_pfx="original_")
    reuse_df = load_sent_df(reuse_corpus, col_pfx="reuse_")

    # Determine sentence pairs
    # TODO: Add support for relevant voyager parameters
    sent_pairs = get_sentence_pairs(
        original_df.get_column("original_text").to_list(),
        reuse_df.get_column("reuse_text").to_list(),
        score_cutoff,
        show_progress_bar=show_progress_bar,
    )

    # Build and save quote pairs if any are found
    if len(sent_pairs):
        quote_pairs = compile_quote_pairs(original_df, reuse_df, sent_pairs)
        # NOTE: Perhaps this should return a DataFrame rather than creating a CSV?
        quote_pairs.write_csv(out_csv)
        logger.info(f"Saved {len(quote_pairs):,} quote pairs to {out_csv}")
    else:
        logger.info(
            f"No sentence pairs for score cutoff = {score_cutoff} ; output file not created."
        )



"""
Preliminary script and method to create sentence corpora from input
files in supported formats.

NOTE: Currently this script can only take single input file

Example Usage:

    `create.py input_text.txt out.csv`

"""

import argparse
import csv
import logging
import pathlib
import sys

from remarx.sentence.corpus.base_input import FileInput
from remarx.utils import configure_logging


def create_corpus(
    input_file: pathlib.Path,
    output_csv: pathlib.Path,
    filename_override: str | None = None,
) -> None:
    """
    Create and save a sentence corpus from the provided input file to the
    provided output path `output_csv`.

    NOTE: An error will be raised if the input file is not a type supported by
    `FileInput`.
    """
    if not input_file.is_file():
        raise ValueError(f"Input file {input_file} does not exist")

    text_input = FileInput.create(input_file, filename_override=filename_override)
    field_names = text_input.field_names

    with output_csv.open(mode="w", newline="") as csvfile:
        csvwriter = csv.DictWriter(csvfile, field_names, extrasaction="ignore")
        csvwriter.writeheader()
        csvwriter.writerows(text_input.get_sentences())


def main() -> None:
    """
    Command-line access to sentence corpus creation for supported input formats
    """
    parser = argparse.ArgumentParser(
        description="Generate a sentence corpus from a supported input file"
    )
    parser.add_argument(
        "input_file",
        type=pathlib.Path,
        help="Path to input file",
    )
    parser.add_argument(
        "output_csv", type=pathlib.Path, help="Path to output sentence corpus (CSV)"
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Verbose output (debug logging)",
        default=False,
    )

    args = parser.parse_args()

    log_level = logging.DEBUG if args.verbose else logging.INFO

    configure_logging(sys.stdout, log_level=log_level)
    create_corpus(
        args.input_file,
        args.output_csv,
    )


if __name__ == "__main__":
    main()



"""
Unit tests for sentence segmentation functionality.
"""

from unittest.mock import Mock, patch

from stanza import DownloadMethod, Pipeline
from stanza.models.common.doc import Document, Sentence

from remarx.sentence.segment import segment_text


def create_mock_sentence(text: str, start_char: int = 0) -> Mock:
    """Helper function: create a mock sentence with the given text and start character."""
    return Mock(spec=Sentence, text=text, tokens=[Mock(start_char=start_char)])


class TestSegmentTextIntoSentences:
    """Test cases for the segment_text_into_sentences function."""

    @patch("remarx.sentence.segment.stanza.Pipeline")
    def test_segment_text_indices(self, mock_pipeline_class: Mock) -> None:
        """Test text segmentation with character indices."""
        # Setup mock
        mock_sentence1 = create_mock_sentence("Erster Satz.", 0)
        mock_sentence2 = create_mock_sentence("Zweiter Satz.", 14)

        mock_doc = Mock(spec=Document, sentences=[mock_sentence1, mock_sentence2])

        mock_pipeline = Mock(spec=Pipeline, return_value=mock_doc)
        mock_pipeline_class.return_value = mock_pipeline

        # Test
        text = "Erster Satz. Zweiter Satz."
        result = segment_text(text)

        # Assertions
        assert len(result) == 2
        assert result[0] == (0, "Erster Satz.")
        assert result[1] == (14, "Zweiter Satz.")

    @patch("remarx.sentence.segment.stanza.Pipeline")
    def test_segment_text_empty_text(self, mock_pipeline_class: Mock) -> None:
        """Test segmentation of empty text."""
        # Setup mock
        mock_doc = Mock(spec=Document, sentences=[])

        mock_pipeline = Mock(spec=Pipeline, return_value=mock_doc)
        mock_pipeline_class.return_value = mock_pipeline

        # Test
        result = segment_text("")

        # Assertions
        assert result == []

    @patch("remarx.sentence.segment.stanza.Pipeline")
    def test_segment_text_language_parameter(self, mock_pipeline_class: Mock) -> None:
        """Test that language parameter works or not."""
        # Setup mock
        mock_sentence = create_mock_sentence("Hallo Welt.", 0)
        mock_doc = Mock(spec=Document, sentences=[mock_sentence])
        mock_pipeline = Mock(spec=Pipeline, return_value=mock_doc)
        mock_pipeline_class.return_value = mock_pipeline

        # Test with explicit "en" language
        segment_text("Hello world.", language="en")
        mock_pipeline_class.assert_called_with(
            lang="en",
            processors="tokenize",
            download_method=DownloadMethod.REUSE_RESOURCES,
        )

        # Reset mock for second test
        mock_pipeline_class.reset_mock()

        # Test with default language (should be "de")
        segment_text("Hallo Welt.")
        mock_pipeline_class.assert_called_with(
            lang="de",
            processors="tokenize",
            download_method=DownloadMethod.REUSE_RESOURCES,
        )



"""remarx: Marx quote identification for the CDH Citing Marx project."""

# Copyright (c) 2025, Center for Digital Humanities, Princeton University
# SPDX-License-Identifier: Apache-2.0

from remarx import app, sentence

__version__ = "0.3rc1"

__all__ = ["__version__", "app", "sentence"]



from remarx.app import corpus_builder, quote_finder
from remarx.app.utils import configure_logging


def test_corpus_builder_app():
    # Run the app - if it crashes, the test will fail
    corpus_builder.app.run()


def test_quote_finder_app():
    # Run the app - if it crashes, the test will fail
    quote_finder.app.run()


def test_corpus_builder_app_logging(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)
    configure_logging()
    corpus_builder.app.run()

    log_files = list((tmp_path / "logs").iterdir())
    log_text = log_files[-1].read_text()
    assert "Remarx Corpus Builder notebook started" in log_text


def test_quote_finder_app_logging(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)
    configure_logging()
    quote_finder.app.run()

    log_files = list((tmp_path / "logs").iterdir())
    log_text = log_files[-1].read_text()
    assert "Remarx Quote Finder notebook started" in log_text



import contextlib
import io
import logging
import sys
from pathlib import Path

import pytest

from remarx.utils import configure_logging


@pytest.fixture(autouse=True)
def reset_logging():
    """
    This fixture forcibly removes all handlers from the root logger before each test,
    guaranteeing that logging.basicConfig in the code under test will always add a fresh handler.
    """
    root_logger = logging.getLogger()
    for handler in root_logger.handlers[:]:
        root_logger.removeHandler(handler)
        with contextlib.suppress(Exception):
            handler.close()


def test_configure_logging_default_creates_timestamped_filename(tmp_path, monkeypatch):
    """Test that the default configuration creates a timestamped log filename and directory."""
    # Run in a temporary CWD so logs land under tmp_path/logs/
    monkeypatch.chdir(tmp_path)
    created_path = configure_logging()

    assert isinstance(created_path, Path)
    logs_dir = created_path.parent
    assert logs_dir == tmp_path / "logs"
    assert logs_dir.is_dir()

    # Check that the log file name starts with "remarx_" and ends with ".log"
    assert created_path.name.startswith("remarx_")
    assert created_path.suffix == ".log"

    # Check root logger level is the expected default (INFO)
    root_logger = logging.getLogger()
    assert root_logger.getEffectiveLevel() == logging.INFO

    # there should be only one handler with our fixture, which should be a FileHandler
    handler = root_logger.handlers[-1]
    assert isinstance(handler, logging.FileHandler)
    assert Path(handler.baseFilename) == created_path

    # Stanza logger level default
    assert logging.getLogger("stanza").getEffectiveLevel() == logging.ERROR


# use parametrize to test with different streams
@pytest.mark.parametrize("stream", [sys.stdout, sys.stderr, io.StringIO()])
def test_configure_logging_stream(tmp_path, monkeypatch, stream):
    # Run in a temporary CWD and ensure no logs/ directory is created when streaming to a text stream
    monkeypatch.chdir(tmp_path)
    created_path = configure_logging(log_destination=stream, log_level=logging.INFO)

    assert created_path is None

    root_logger = logging.getLogger()
    handler = root_logger.handlers[-1]
    assert isinstance(handler, logging.StreamHandler)
    assert getattr(handler, "stream", None) is stream

    # Confirm that no log directory or file was created as we logged to a stream
    assert not (tmp_path / "logs").exists()


def test_configure_logging_specific_file(tmp_path):
    target_path = tmp_path / "nested" / "custom.log"
    created_path = configure_logging(target_path, log_level=logging.DEBUG)

    assert created_path == target_path
    # Only require that the parent directory exists; file may be created on first write
    assert target_path.parent.exists()

    root_logger = logging.getLogger()
    handler = root_logger.handlers[-1]
    assert isinstance(handler, logging.FileHandler)
    assert Path(handler.baseFilename) == target_path
    # The handler should be set to the correct log level (DEBUG or NOTSET if inherited)
    assert handler.level in (logging.NOTSET, logging.DEBUG)


def test_configure_logging_with_stanza_log_level(tmp_path, monkeypatch):
    # Use a clean temp directory for logs
    monkeypatch.chdir(tmp_path)
    configure_logging(stanza_log_level=logging.DEBUG)

    # Check that the stanza logger is set to DEBUG
    stanza_logger = logging.getLogger("stanza")
    assert stanza_logger.getEffectiveLevel() == logging.DEBUG



import pathlib
from collections.abc import Generator
from unittest.mock import Mock, patch

from remarx.sentence.corpus.base_input import FileInput
from remarx.sentence.corpus.text_input import TextInput


def test_init(tmp_path: pathlib.Path):
    txt_file = tmp_path / "input.txt"
    txt_input = TextInput(input_file=txt_file)
    assert txt_input.input_file == txt_file


def test_file_name(tmp_path: pathlib.Path):
    txt_filename = "my_input.txt"
    txt_file = tmp_path / txt_filename
    txt_input = TextInput(input_file=txt_file)
    assert txt_input.file_name == txt_filename


def test_field_names(tmp_path: pathlib.Path):
    assert TextInput.field_names == FileInput.field_names


def test_get_text(tmp_path: pathlib.Path):
    txt_file = tmp_path / "input.txt"
    text_contents = "placeholder content"
    txt_file.write_text(text_contents)

    txt_input = TextInput(input_file=txt_file)
    text_result = txt_input.get_text()
    # expect a generator with one item, with the content added to the file
    assert isinstance(text_result, Generator)
    text_result = list(text_result)
    assert len(text_result) == 1
    first_result = next(iter(text_result))
    assert isinstance(first_result, dict)
    # text content matches expected results
    assert first_result["text"] == text_contents
    # dict should only includes text
    assert list(first_result.keys()) == ["text"]


def simple_segmenter(text: str):
    # for testing purposes, dummy segmenter that splits input text in half
    half_text_len = int(len(text) / 2)
    # segment text returns a generator of tuple of character index, sentence text
    return ((0, text[:half_text_len]), (half_text_len, text[half_text_len:]))


@patch("remarx.sentence.corpus.base_input.segment_text")
def test_get_sentences(mock_segment_text: Mock, tmp_path: pathlib.Path):
    txt_file = tmp_path / "input.txt"
    text_content = "more placeholder content"
    txt_file.write_text(text_content)
    # call simple segmenter to split input text in two
    mock_segment_text.side_effect = simple_segmenter

    txt_input = TextInput(input_file=txt_file)
    sentences = txt_input.get_sentences()
    # expect a generator with two item, with the content added to the file
    assert isinstance(sentences, Generator)
    # consume the generator
    sentences = list(sentences)
    assert len(sentences) == 2

    # expect segmentation method to be called only once
    mock_segment_text.assert_called_once()

    first_sentence = sentences[0]
    assert isinstance(first_sentence, dict)
    # should not be the full text content
    assert first_sentence["text"] != text_content
    # should _start_ with the text content
    assert first_sentence["text"].startswith(text_content[:5])
    assert first_sentence["file"] == txt_file.name
    assert first_sentence["sent_index"] == 0

    second_sentence = sentences[1]
    # should not be the full text content
    assert second_sentence["text"] != text_content
    # but should _end_ with the text content
    assert second_sentence["text"].endswith(text_content[-5:])
    assert second_sentence["file"] == txt_file.name



import logging
from unittest.mock import Mock, patch

import marimo
import pytest
from marimo._server.asgi import ASGIAppBuilder

# Types pulled from marimo source code
from starlette.types import ASGIApp

import remarx
from remarx.app.utils import (
    create_header,
    create_temp_input,
    get_current_log_file,
    launch_app,
    lifespan,
    redirect_root,
)


@patch("remarx.app.utils.uvicorn.run")
@patch("remarx.app.utils.mo")
@patch("remarx.app.utils.configure_logging")
def test_launch_app(mock_configure_logging, mock_mo, mock_uvicorn_run):
    """Test that launch_app sets up the application structure correctly"""
    mock_server = Mock(spec=ASGIAppBuilder)
    mock_mo.create_asgi_app.return_value = mock_server
    mock_server.with_app.return_value = mock_server
    mock_server.build.return_value = Mock(spec=ASGIApp)

    # Mock configure_logging to return a log file path
    mock_configure_logging.return_value = "/path/to/log/file.log"

    launch_app()

    # Verify logging is configured first
    mock_configure_logging.assert_called_once()

    # Verify marimo app creation and configuration
    mock_mo.create_asgi_app.assert_called_once()
    assert mock_server.with_app.call_count == 2  # corpus-builder and quote-finder
    mock_server.build.assert_called_once()

    # Verify server startup
    mock_uvicorn_run.assert_called_once()


@patch("remarx.app.utils.FileUploadResults")
def test_create_temp_input(mock_upload):
    """Test temporary file creation and cleanup"""
    # Create mock file upload
    mock_upload.name = "file.txt"
    mock_upload.contents = b"bytes"

    # Normal case
    with create_temp_input(mock_upload) as tf:
        assert tf.is_file()
        assert tf.suffix == ".txt"
        assert tf.read_text() == "bytes"
    assert not tf.is_file()

    # Check temp file is closed if an exception is raised
    try:
        with create_temp_input(mock_upload) as tf:
            raise ValueError
    except ValueError:
        # catch thrown exception
        pass
    assert not tf.is_file()


@patch("remarx.app.utils.mo.vstack")
@patch("remarx.app.utils.mo.nav_menu")
@patch("remarx.app.utils.mo.md")
def test_create_header(mock_md, mock_nav_menu, mock_vstack):
    """Test create_header function"""
    # Mock the marimo functions
    ## Work around for HTML centering
    mock_md.side_effect = (
        lambda x: x
        if x == "---"
        else Mock(spec=marimo.Html, **{"text": x, "center.return_value": x})
    )
    mock_nav_menu.return_value = Mock(
        spec=marimo.Html, **{"center.return_value": "nav_bar"}
    )
    mock_vstack.return_value = "mocked_header"

    result = create_header()
    mock_nav_menu.assert_called_once()
    assert mock_md.call_count == 4
    mock_vstack.assert_called_once_with(
        [
            "# `remarx`",
            f"Running version: {remarx.__version__}",
            "---",
            "nav_bar",
            "---",
        ]
    )
    assert result == "mocked_header"


@patch("remarx.app.utils.webbrowser")
@pytest.mark.asyncio
async def test_lifespan(mock_webbrowser):
    """Test lifespan context manager"""
    from fastapi import FastAPI

    # Create a mock app
    app = FastAPI()

    # Test the lifespan context manager
    async with lifespan(app):
        mock_webbrowser.open.assert_called_once_with("http://localhost:8000/")


@pytest.mark.asyncio
async def test_redirect_root():
    """Test redirect_root function directly"""
    response = await redirect_root()

    assert response.status_code == 302
    assert response.headers["location"] == "/corpus-builder"


def test_get_current_log_file(tmp_path):
    logger = logging.getLogger()
    original_handlers = logger.handlers[:]

    # remove all handlers
    for handler in original_handlers:
        logger.removeHandler(handler)

    # should return None when there is no FileHandler
    stream_handler = logging.StreamHandler()
    logger.addHandler(stream_handler)
    try:
        result = get_current_log_file()
        assert result is None
    finally:
        logger.removeHandler(stream_handler)

    # should return log_file path when there is a FileHandler
    log_file = tmp_path / "test.log"
    file_handler = logging.FileHandler(log_file)
    logger.addHandler(file_handler)
    try:
        result = get_current_log_file()
        assert result == log_file
    finally:
        logger.removeHandler(file_handler)
        file_handler.close()

    # restore original handlers
    for handler in original_handlers:
        logger.addHandler(handler)


def test_launch_app_logging(tmp_path, monkeypatch):
    """Test that launch_app logs the correct startup messages"""
    monkeypatch.chdir(tmp_path)

    with patch("remarx.app.utils.uvicorn.run"):
        launch_app()

    log_files = list((tmp_path / "logs").iterdir())
    log_text = log_files[-1].read_text()
    assert "Remarx application starting" in log_text
    assert "Logs are being written to:" in log_text



"""
Functionality related to parsing ALTO XML content packaged within a zipfile,
with the goal of creating a sentence corpus with associated metadata from ALTO.
"""

import logging
import pathlib
from collections.abc import Generator
from dataclasses import dataclass
from functools import cached_property
from timeit import default_timer as time
from typing import ClassVar
from zipfile import ZipFile

from lxml import etree
from natsort import natsorted
from neuxml import xmlmap

from remarx.sentence.corpus.base_input import FileInput, SectionType

logger = logging.getLogger(__name__)


ALTO_NAMESPACE_V4: str = "http://www.loc.gov/standards/alto/ns-v4#"


class AltoXmlObject(xmlmap.XmlObject):
    """
    Base :class:`neuxml.xmlmap.XmlObject` class for ALTO-XML content.
    """

    # alto namespace v4; we may eventually need to support other versions
    ROOT_NAMESPACES: ClassVar[dict[str, str]] = {"alto": ALTO_NAMESPACE_V4}


class AltoBlock(AltoXmlObject):
    """
    Base class for an ALTO element with position information.
    """

    vertical_position = xmlmap.FloatField("@VPOS")
    horizontal_position = xmlmap.FloatField("@HPOS")


class TextLine(AltoBlock):
    """
    Single line of text (`TextLine`) in an ALTO document
    """

    text_content = xmlmap.StringField("alto:String/@CONTENT")

    def __str__(self) -> str:
        """
        Override default string method to return text content of this line.
        """
        return self.text_content


class TextBlock(AltoBlock):
    """
    Block of text with one or more lines.
    """

    lines = xmlmap.NodeListField("alto:TextLine", TextLine)

    @cached_property
    def sorted_lines(self) -> list[TextLine]:
        """
        Returns a list of TextLines for this block, sorted by vertical position.
        """
        # there's no guarantee that xml document order follows page order,
        # so sort by @VPOS (may need further refinement for more complicated layouts)
        return sorted(self.lines, key=lambda line: line.vertical_position)

    @property
    def text_content(self) -> str:
        """
        Text contents of this block; newline-delimited content of
        each line within this block, sorted by vertical position.
        """
        return "\n".join([line.text_content for line in self.sorted_lines])


class AltoDocument(AltoXmlObject):
    """
    :class:`neuxml.xmlmap.XmlObject` instance for a single ALTO XML file
    """

    blocks = xmlmap.NodeListField(".//alto:TextBlock", TextBlock)
    lines = xmlmap.NodeListField(".//alto:TextLine", TextLine)

    def is_alto(self) -> bool:
        """
        Check if this is an ALTO-XML document, based on the root element
        """
        # parse with QName to access namespace and tag name without namespace
        root_element = etree.QName(self.node.tag)
        # both must match
        return (
            root_element.namespace == ALTO_NAMESPACE_V4
            and root_element.localname == "alto"
        )

    @property
    def sorted_blocks(self) -> list[TextBlock]:
        """
        Returns a list of TextBlocks for this page, sorted by vertical position.
        """
        # there's no guarantee that xml document order follows page order,
        # so sort by @VPOS (may need further refinement for more complicated layouts).
        # NOTE: in some cases, a textblock may not have a VPOS attribute;
        # in that case, use the position for the first line
        # (text block id = eSc_dummyblock_, but appears to have real content)
        # if block has no line, sort text block last
        if not self.blocks:
            return []
        return sorted(
            self.blocks,
            key=lambda block: block.vertical_position
            or (
                block.sorted_lines[0].vertical_position if block.lines else float("inf")
            ),
        )

    def text_chunks(self) -> Generator[dict[str, str]]:
        """
        Returns a generator of a dictionary of text content and section type,
        one dictionary per text block on the page.
        """
        # yield by block, since in future we may set section type
        # based on block-level semantic tagging
        for block in self.sorted_blocks:
            yield {"text": block.text_content, "section_type": SectionType.TEXT.value}


@dataclass
class ALTOInput(FileInput):
    """
    Preliminary FileInput implementation for ALTO XML delivered as a zipfile.
    Iterates through ALTO XML members and stubs out chunk yielding for future parsing.
    """

    field_names: ClassVar[tuple[str, ...]] = (*FileInput.field_names, "section_type")
    "List of field names for sentences originating from ALTO XML content."

    file_type: ClassVar[str] = ".zip"
    "Supported file extension for ALTO zipfiles (.zip)"

    def get_text(self) -> Generator[dict[str, str], None, None]:
        """
        Iterate over ALTO XML files contained in the zipfile and return
        a generator of text content.
        """
        num_files = 0
        num_valid_files = 0

        start = time()
        with ZipFile(self.input_file) as archive:
            # iterate over all files in the zipfile;
            # use natural sorting to process in logical order
            for zip_filepath in natsorted(archive.namelist()):
                num_files += 1
                base_filename = pathlib.Path(zip_filepath).name
                # ignore & log non-xml files
                if not base_filename.lower().endswith(".xml"):
                    logger.info(
                        f"Ignoring non-xml file included in ALTO zipfile: {zip_filepath}"
                    )
                    continue
                # if the file is .xml, attempt to open as an ALTO XML
                with archive.open(zip_filepath) as xmlfile:
                    logger.info(f"Processing XML file {zip_filepath}")
                    # zipfile archive open returns a file-like object
                    try:
                        alto_xmlobj = xmlmap.load_xmlobject_from_file(
                            xmlfile, AltoDocument
                        )
                    except etree.XMLSyntaxError as err:
                        logger.warning(f"Skipping {zip_filepath} : invalid XML")
                        logger.debug(f"XML syntax error: {err}", exc_info=err)
                        continue

                if not alto_xmlobj.is_alto():
                    # TODO: add unit test for this case
                    logger.warning(
                        f"Skipping non-ALTO XML file {zip_filepath} (root element {alto_xmlobj.node.tag})"
                    )
                    continue

                num_valid_files += 1
                # report total # blocks, lines for each file as processed
                logger.debug(
                    f"{base_filename}: {len(alto_xmlobj.blocks)} blocks, {len(alto_xmlobj.lines)} lines"
                )

                # use the base xml file as filename here, rather than zipfile for all
                for chunk in alto_xmlobj.text_chunks():
                    yield chunk | {"file": base_filename}

                # warn if a document has no lines
                if len(alto_xmlobj.lines) == 0:
                    logger.warning(
                        f"No text lines found in ALTO XML file: {base_filename}"
                    )

        elapsed_time = time() - start
        logger.info(
            f"Processed {self.file_name} with {num_files} files ({num_valid_files} valid ALTO) in {elapsed_time:.1f} seconds"
        )

        # error if no valid files were found
        if num_valid_files == 0:
            raise ValueError(f"No valid ALTO XML files found in {self.file_name}")



"""
Base file input class with common functionality. Provides a factory
method for initialization of known input classes based on supported
file types.

To initialize the appropriate subclass for a supported file type,
use [FileInput.create()][remarx.sentence.corpus.base_input.FileInput.create].

For a list of supported file types across all registered input classes,
use [FileInput.supported_types()][remarx.sentence.corpus.base_input.FileInput.supported_types].

Subclasses must define a supported `file_type` extension and implement
the `get_text` method. For discovery, input classes must be imported in
`remarx.sentence.corpus.__init__` and included in `__all__` to ensure
they are found as available input classes.

"""

import pathlib
from collections.abc import Generator
from dataclasses import dataclass
from enum import StrEnum
from functools import cached_property
from typing import Any, ClassVar, Self

from remarx.sentence.segment import segment_text


@dataclass
class FileInput:
    """Base class for file input for sentence corpus creation"""

    input_file: pathlib.Path
    "Reference to input file. Source of content for sentences."

    filename_override: str = None
    "Optional filename override, e.g. when using temporary files as input"

    field_names: ClassVar[tuple[str, ...]] = ("sent_id", "file", "sent_index", "text")
    "List of field names for sentences from text input files."

    file_type: ClassVar[str]
    "Supported file extension; subclasses must define"

    @cached_property
    def file_name(self) -> str:
        """
        Input file name. Associated with sentences in generated corpus.
        """
        return self.filename_override or self.input_file.name

    def get_text(self) -> Generator[dict[str, str]]:
        """
        Get plain-text contents for this input file with any desired chunking
        (e.g. pages or other semantic unit).
        Subclasses must implement; no default implementation.

        :returns: Generator with a dictionary of text and any other metadata
        that applies to this unit of text.
        """
        raise NotImplementedError

    def get_extra_metadata(
        self, chunk_info: dict[str, Any], _char_idx: int, sentence: str
    ) -> dict[str, Any]:
        """
        Hook method for subclasses to override to provide extra metadata for a sentence (e.g. line number).

        :returns: Dictionary of additional metadata fields to include, or empty dict
        """
        return {}

    def get_sentences(self) -> Generator[dict[str, Any]]:
        """
        Get sentences for this file, with associated metadata.

        :returns: Generator of one dictionary per sentence; dictionary
        always includes: `text` (text content), `file` (filename),
        `sent_index` (sentence index within the document), and `sent_id`
        (sentence id). It may include other metadata, depending
        on the input file type.
        """
        # zero-based sentence index for this file, across all chunks
        sentence_index = 0
        for chunk_info in self.get_text():
            # each chunk of text is a dictionary that at minimum
            # contains text for that chunk; it may include other metadata
            chunk_text = chunk_info["text"]
            for _char_idx, sentence in segment_text(chunk_text):
                # for each sentence, yield text, filename, and sentence index
                # with any other metadata included in chunk_info

                # character index is not included in output,
                # but may be useful for sub-chunk metadata (e.g., line number)

                # specify input file name first;
                # chunk-specific filename take precedence (e.g. alto file within zip)
                yield (
                    {"file": self.file_name}
                    | chunk_info
                    | {
                        "text": sentence,
                        "sent_index": sentence_index,
                        "sent_id": f"{self.file_name}:{sentence_index}",
                    }
                    # Include any extra metadata (subclass specific)
                    | self.get_extra_metadata(chunk_info, _char_idx, sentence)
                )

                # increment sentence index
                sentence_index += 1

    @classmethod
    def subclasses(cls) -> list[type[Self]]:
        """
        List of available file input classes.
        """
        return cls.__subclasses__()

    @classmethod
    def subclass_by_type(cls) -> dict[str, type[Self]]:
        """
        Dictionary of subclass by supported file extension for available
        input classes.
        """
        return {subcls.file_type: subcls for subcls in cls.subclasses()}

    @classmethod
    def supported_types(cls) -> list[str]:
        """
        Unique list of supported file extensions for available input classes.
        """
        return list({subcls.file_type for subcls in cls.subclasses()})

    @classmethod
    def create(
        cls, input_file: pathlib.Path, filename_override: str | None = None
    ) -> Self:
        """
        Instantiate and return the appropriate input class for the specified
        input file.  Takes an optional filename override parameter,
        which is passed through to the input class.

        :raises ValueError: if input_file is not a supported type
        """
        input_cls = cls.subclass_by_type().get(input_file.suffix.lower())
        # for now, check based on file extension
        # NOTE: this will change when we add support for METS-ALTO
        if input_cls is None:
            # include supported types in error to aid debugging
            # sort so display order is reliable
            supported_types = ", ".join(sorted(cls.supported_types()))
            raise ValueError(
                f"{input_file.suffix} is not a supported input type (must be one of {supported_types})"
            )
        return input_cls(input_file=input_file, filename_override=filename_override)


class SectionType(StrEnum):
    """Section types declaration, for distinguishing different text sections within corpus documents."""

    TEXT = "text"
    """Body text content from the document."""

    FOOTNOTE = "footnote"
    """Footnote content extracted from the document."""



import pathlib
from unittest.mock import patch

import pytest

from remarx.sentence.corpus.base_input import FileInput


def test_subclasses():
    # check that expected input subclasses are found
    subclass_names = [cls.__name__ for cls in FileInput.subclasses()]
    # NOTE: that we use names here rather than importing, to
    # confirm subclasses are found without a direct import
    for input_cls_name in ["TextInput", "TEIinput", "ALTOInput"]:
        assert input_cls_name in subclass_names


def test_init(tmp_path: pathlib.Path):
    txt_file = tmp_path / "input.txt"
    txt_input = FileInput(input_file=txt_file)
    assert txt_input.input_file == txt_file


def test_file_name(tmp_path: pathlib.Path):
    txt_filename = "my_input.txt"
    txt_file = tmp_path / txt_filename
    txt_input = FileInput(input_file=txt_file)
    assert txt_input.file_name == txt_filename


def test_file_name_override(tmp_path: pathlib.Path):
    real_txt_filename = "my_input.txt"
    txt_file = tmp_path / "tmp_abc_input_foo.txt"
    txt_input = FileInput(input_file=txt_file, filename_override=real_txt_filename)
    assert txt_input.file_name == real_txt_filename


def test_field_names(tmp_path: pathlib.Path):
    assert FileInput.field_names == ("sent_id", "file", "sent_index", "text")


def test_supported_types():
    # check for expected supported types
    # NOTE: checking directly to avoid importing input classes
    assert set(FileInput.supported_types()) == {".txt", ".xml", ".zip"}


def test_get_text(tmp_path: pathlib.Path):
    # get text is not implemented in the base class
    txt_file = tmp_path / "test.txt"
    base_input = FileInput(input_file=txt_file)
    with pytest.raises(NotImplementedError):
        base_input.get_text()


@patch("remarx.sentence.corpus.base_input.segment_text")
@patch.object(FileInput, "get_text")
def test_get_sentences(mock_text, mock_segment, tmp_path: pathlib.Path):
    mock_segment.side_effect = lambda x: [(0, x)]
    mock_text.return_value = [{"id": i, "text": f"s{i}"} for i in range(3)]
    txt_file = tmp_path / "test.txt"
    base_input = FileInput(input_file=txt_file)

    results = list(base_input.get_sentences())
    assert len(results) == 3
    for i in range(3):
        assert results[i] == {
            "id": i,
            "text": f"s{i}",
            "file": "test.txt",
            "sent_index": i,
            "sent_id": f"test.txt:{i}",
        }


@patch("remarx.sentence.corpus.base_input.segment_text")
@patch.object(FileInput, "get_text")
def test_get_sentences_chunk_filename(mock_text, mock_segment, tmp_path: pathlib.Path):
    mock_segment.side_effect = lambda x: [(0, x)]

    mock_text.return_value = [
        {"id": i, "text": f"s{i}", "file": f"page_{i}.txt"} for i in range(3)
    ]
    txt_file = tmp_path / "test.txt"
    base_input = FileInput(input_file=txt_file)

    results = list(base_input.get_sentences())
    for i in range(3):
        # chunk info filename should take precedence if specified
        assert results[i]["file"] == f"page_{i}.txt"


def test_create_txt(tmp_path: pathlib.Path):
    from remarx.sentence.corpus.text_input import TextInput

    txt_file = tmp_path / "input.txt"
    txt_input = FileInput.create(input_file=txt_file)
    assert isinstance(txt_input, TextInput)


def test_create_exts(tmp_path: pathlib.Path):
    from remarx.sentence.corpus.text_input import TextInput

    txt_file = tmp_path / "upper.TXT"
    txt_input = FileInput.create(input_file=txt_file)
    assert isinstance(txt_input, TextInput)

    txt_file = tmp_path / "mixed.TxT"
    txt_input = FileInput.create(input_file=txt_file)
    assert isinstance(txt_input, TextInput)


def test_create_filename_override(tmp_path: pathlib.Path):
    from remarx.sentence.corpus.text_input import TextInput

    txt_file = tmp_path / "tmp_foo_bar_input.txt"
    real_filename = "input.txt"
    txt_input = FileInput.create(input_file=txt_file, filename_override=real_filename)
    assert isinstance(txt_input, TextInput)
    assert txt_input.file_name == real_filename


@patch("remarx.sentence.corpus.tei_input.TEIDocument")
def test_create_tei(mock_tei_doc, tmp_path: pathlib.Path):
    from remarx.sentence.corpus.tei_input import TEIinput

    xml_input_file = tmp_path / "input.xml"
    xml_input = FileInput.create(input_file=xml_input_file)
    assert isinstance(xml_input, TEIinput)
    mock_tei_doc.init_from_file.assert_called_with(xml_input_file)


def test_create_alto(tmp_path: pathlib.Path):
    from remarx.sentence.corpus.alto_input import ALTOInput

    zip_input_file = tmp_path / "input.zip"
    zip_input_file.touch()
    zip_input = FileInput.create(input_file=zip_input_file)
    assert isinstance(zip_input, ALTOInput)


def test_create_unsupported(tmp_path: pathlib.Path):
    test_file = tmp_path / "input.test"
    with pytest.raises(
        ValueError,
        match="\\.test is not a supported input type \\(must be one of \\.txt, \\.xml, \\.zip\\)",
    ):
        FileInput.create(input_file=test_file)



"""
Utility methods associated with the remarx app
"""

import contextlib
import logging
import pathlib
import tempfile
import webbrowser
from collections.abc import AsyncGenerator, Generator
from contextlib import asynccontextmanager

import marimo as mo
import uvicorn
from fastapi import FastAPI
from fastapi.responses import RedirectResponse

# Does this class have a public facing type definition?
from marimo._plugins.ui._impl.input import FileUploadResults

import remarx
from remarx.utils import configure_logging

# Server configuration
HOST = "localhost"
PORT = 8000


@asynccontextmanager
async def lifespan(app: FastAPI) -> AsyncGenerator[None, None]:
    """Lifespan context manager to open browser when server starts"""
    webbrowser.open(f"http://{HOST}:{PORT}/")
    yield


async def redirect_root() -> RedirectResponse:
    """Redirect root path to corpus-builder, since app currently has no home page"""
    return RedirectResponse(url="/corpus-builder", status_code=302)


def launch_app() -> None:
    """Launch the remarx app in the default web browser"""
    # Configure logging once at startup to prevent multiple log files created by the notebooks
    log_file_path = configure_logging()

    # Log that the app is starting
    logger = logging.getLogger("remarx-app")
    logger.info("Remarx application starting")
    if log_file_path:
        logger.info(f"Logs are being written to: {log_file_path}")

    # Create marimo asgi app
    server = mo.create_asgi_app()

    # Add notebooks
    server = server.with_app(
        path="/corpus-builder", root=remarx.app.corpus_builder.__file__
    )
    server = server.with_app(
        path="/quote-finder", root=remarx.app.quote_finder.__file__
    )

    # Create a FastAPI app with lifespan to open browser
    app = FastAPI(lifespan=lifespan)

    # Add redirect from root to corpus-builder
    app.get("/")(redirect_root)

    app.mount("/", server.build())

    # Run server
    uvicorn.run(app, host=HOST, port=PORT)


def get_current_log_file() -> pathlib.Path | None:
    """
    Get the path to the current log file from the root logger's file handler.
    Returns None if logging is not configured to a file.
    """
    root_logger = logging.getLogger()
    for handler in root_logger.handlers:
        if isinstance(handler, logging.FileHandler):
            return pathlib.Path(handler.baseFilename)
    return None


def create_header() -> None:
    """Create the header for the remarx notebooks"""
    return mo.vstack(
        [
            mo.md("# `remarx`").center(),
            mo.md(f"Running version: {remarx.__version__}").center(),
            mo.md("---"),
            mo.nav_menu(
                {
                    "/corpus-builder": "### Sentence Corpus Builder",
                    "/quote-finder": "### Quote Finder",
                }
            ).center(),
            mo.md("---"),
        ]
    )


@contextlib.contextmanager
def create_temp_input(
    file_upload: FileUploadResults,
) -> Generator[pathlib.Path, None, None]:
    """
    Context manager to create a temporary file with the file contents and name of a file uploaded
    to a web browser as returned by marimo.ui.file. This should be used in with statements.

    :returns: Yields the path to the temporary file
    """
    temp_file = tempfile.NamedTemporaryFile(  # noqa: SIM115
        delete=False,
        suffix=pathlib.Path(file_upload.name).suffix,
    )
    try:
        temp_file.write(file_upload.contents)
        # Close to ensure write occurs
        temp_file.close()
        yield pathlib.Path(temp_file.name)
    finally:
        if not temp_file.closed:
            temp_file.close()
        pathlib.Path.unlink(temp_file.name)



"""
The marimo notebook corresponding to the `remarx` application. The application
can be launched by running the command `remarx-app` or via marimo.

Example Usage:

    `remarx-app`

    `marimo run app.py`
"""

import marimo

__generated_with = "0.15.2"
app = marimo.App(width="medium")


@app.cell
def _():
    import csv
    import marimo as mo
    import pathlib
    import tempfile

    import logging
    import remarx
    from remarx.app.utils import create_header, create_temp_input, get_current_log_file
    from remarx.sentence.corpus import FileInput
    from remarx.quotation.pairs import find_quote_pairs
    return (
        FileInput,
        create_header,
        create_temp_input,
        find_quote_pairs,
        get_current_log_file,
        logging,
        mo,
        pathlib,
        remarx,
    )


@app.cell
def _(create_header):
    create_header()
    return


@app.cell
def _(get_current_log_file, logging):
    # Get log file path from already configured logging
    log_file_path = get_current_log_file()

    # Log that UI started
    logger = logging.getLogger("remarx-app")
    logger.info("Remarx Quote Finder notebook started")

    return (log_file_path,)


@app.cell
def _(mo):
    mo.md(
        r"""
    ## :mag: Quotation Finder
    Determine and identify the passages of a text corpus (**reuse**) that quote passages from texts in another corpus (**original**).
    This process requires sentence corpora (`CSVs`) created in the previous section.
    """
    )
    return


@app.cell
def _(mo):
    mo.md(
        r"""
    ### 1. Select Input CSV Files

    Browse and select CSV files for each category (currently only supports one file each):

    - **Original Sentence Corpora**: Sentence-level text corpora of the texts that we are searching for quotations of.
    - **Reuse Sentence Corpora**: Text that may contain quotations from the original text that will be detected.
    """
    )
    return


@app.cell
def _(mo, pathlib):
    # Create file browsers for quotation detection (CSV files only)
    original_csv_browser = mo.ui.file_browser(
        selection_mode="file",
        multiple=True,
        initial_path=pathlib.Path.home(),
        filetypes=[".csv"],
    )

    reuse_csv_browser = mo.ui.file_browser(
        selection_mode="file",
        multiple=True,
        initial_path=pathlib.Path.home(),
        filetypes=[".csv"],
    )
    return original_csv_browser, reuse_csv_browser


@app.cell
def _(mo, original_csv_browser, reuse_csv_browser):
    # Process file selections for quotation detection
    original_csvs = original_csv_browser.value or []
    reuse_csvs = reuse_csv_browser.value or []

    original_msg = (
        f"{len(original_csvs)} files selected"
        if original_csvs
        else "No original text files selected"
    )
    reuse_msg = (
        f"{len(reuse_csvs)} files selected"
        if reuse_csvs
        else "No reuse text files selected"
    )

    original_callout_type = "success" if original_csvs else "warn"
    reuse_callout_type = "success" if reuse_csvs else "warn"

    # Create side-by-side file browser interface
    mo.hstack(
        [
            mo.callout(
                mo.vstack(
                    [
                        mo.md(
                            "**:card_file_box: Select Original Sentence Corpora (CSVs)**"
                        ).center(),
                        original_csv_browser,
                        mo.md(original_msg),
                    ]
                ),
                kind=original_callout_type,
            ),
            mo.callout(
                mo.vstack(
                    [
                        mo.md(
                            "**:recycle: Select Reuse Sentence Corpora (CSVs)**"
                        ).center(),
                        reuse_csv_browser,
                        mo.md(reuse_msg),
                    ]
                ),
                kind=reuse_callout_type,
            ),
        ],
        widths="equal",
        gap=1.2,
    )
    return original_csvs, reuse_csvs


@app.cell
def _(mo):
    mo.md(
        r"""
    ### 2. Select Output Location

    Select the folder where the resulting quote pairs file should be saved.
    The output CSV file will be named based on the input files.

    *To select a folder, click the file icon to the left of the folder's name.
    A checkmark will appear when a selection is made.
    Clicking anywhere else within the folder's row will cause the browser to navigate to this folder and subsequently display any folders *within* this folder.*
    """
    )
    return


@app.cell
def _(mo, pathlib):
    select_output_dir = mo.ui.file_browser(
        selection_mode="directory",
        multiple=False,
        initial_path=pathlib.Path.home(),
        filetypes=[],  # only show directories
    )
    return (select_output_dir,)


@app.cell
def _(mo, select_output_dir):
    output_dir = select_output_dir.value[0] if select_output_dir.value else None
    output_dir_msg = f"`{output_dir.path}`" if output_dir else "None selected"
    out_callout_type = "success" if output_dir else "warn"

    mo.callout(
        mo.vstack(
            [
                select_output_dir,
                mo.md(f"**Save Location:** {output_dir_msg}"),
            ],
        ),
        kind=out_callout_type,
    )
    return (output_dir,)


@app.cell
def _(mo):
    mo.md(
        r"""
    ### 3. Find Quote Pairs

    Click the "Find Quote Pairs" to run quote detection.
    The quote pairs for the input corpora will be saved as a CSV in the selected save location.
    This output file will be named based on the selected input files.
    """
    )
    return


@app.cell
def _(mo, original_csvs, reuse_csvs, output_dir):
    # Determine inputs based on file & folder selections
    original_file = original_csvs[0] if original_csvs else None
    reuse_file = reuse_csvs[0] if reuse_csvs else None

    output_csv = None
    if original_file and reuse_file and output_dir:
        output_filename = f"quote_pairs_{original_file.path.stem}_{reuse_file.path.stem}.csv"
        output_csv = output_dir.path / output_filename

    original_file_msg = (
        f"`{original_file.path.name}`" if original_file else "*Please select an original corpus file*"
    )

    reuse_file_msg = (
        f"`{reuse_file.path.name}`" if reuse_file else "*Please select a reuse corpus file*"
    )

    dir_msg = (
        f"`{output_dir.path}`"
        if output_dir
        else "*Please select a save location*"
    )

    button = mo.ui.run_button(
        disabled=not (original_file and reuse_file and output_dir),
        label="Find Quote Pairs",
        tooltip="Click to find quote pairs",
    )

    mo.callout(
        mo.vstack(
            [
                mo.md(
                    f"""#### User Selections
                - **Original Corpus:** {original_file_msg}
                - **Reuse Corpus:** {reuse_file_msg}
                - **Save Location:** {dir_msg}
            """
                ),
                button,
            ]
        ),
    )
    return button, output_csv, original_file, reuse_file


@app.cell
def _(button, find_quote_pairs, mo, original_file, reuse_file, output_csv):
    # Find Quote Pairs
    finding_msg = 'Click "Find Quote Pairs" button to start'

    if button.value:
        spinner_msg = f"Finding quote pairs between {original_file.path.name} and {reuse_file.path.name}"
        with mo.status.spinner(title=spinner_msg) as _spinner:
            find_quote_pairs(
                original_corpus=original_file.path,
                reuse_corpus=reuse_file.path,
                out_csv=output_csv,
                show_progress_bar=False
            )
        finding_msg = f":white_check_mark: Quote pairs saved to: {output_csv}"

    mo.md(finding_msg).center()
    return


@app.cell
def _(mo, log_file_path):
    mo.md(f"Logs are being written to: {log_file_path}")
    return


if __name__ == "__main__":
    app.run()



"""
Tests for sentence embedding functionality.
"""

import logging
import re
from unittest.mock import Mock, patch

from remarx.quotation.embeddings import get_sentence_embeddings


@patch("remarx.quotation.embeddings.SentenceTransformer")
def test_get_sentence_embeddings(mock_transformer_class, caplog):
    """Test sentence embedding generation from list of sentences."""

    # Mock the sentence transformer
    mock_model = Mock()
    mock_embeddings = "mock_embeddings"
    mock_model.encode.return_value = mock_embeddings
    mock_transformer_class.return_value = mock_model

    sentences = ["Test sentence 1", "Test sentence 2"]

    result = get_sentence_embeddings(sentences)

    # Verify the model was initialized with default model name
    mock_transformer_class.assert_called_once_with(
        "paraphrase-multilingual-mpnet-base-v2"
    )

    # Verify encode was called with correct parameters
    mock_model.encode.assert_called_once_with(
        sentences,
        normalize_embeddings=True,
        show_progress_bar=False,
    )

    assert result == mock_embeddings

    # Test logging
    caplog.clear()
    with caplog.at_level(logging.INFO):
        result = get_sentence_embeddings(sentences)
    assert len(caplog.record_tuples) == 1
    log = caplog.record_tuples[0]
    assert log[0] == "remarx.quotation.embeddings"
    assert log[1] == logging.INFO
    assert re.fullmatch(
        rf"Generated {len(mock_embeddings)} embeddings in \d+\.\d seconds", log[2]
    )

    # Test with custom model
    mock_transformer_class.reset_mock()
    custom_model = "paraphrase-multilingual-mpnet-base-v3"

    result = get_sentence_embeddings(sentences, model_name=custom_model)

    # Verify custom model was used
    mock_transformer_class.assert_called_once_with(custom_model)
    assert result == mock_embeddings



"""
Functionality for loading and chunking input files for sentence corpus creation.
"""

from remarx.sentence.corpus.alto_input import ALTOInput
from remarx.sentence.corpus.base_input import FileInput
from remarx.sentence.corpus.tei_input import TEI_TAG, TEIDocument, TEIinput, TEIPage
from remarx.sentence.corpus.text_input import TextInput

__all__ = [
    "TEI_TAG",
    "ALTOInput",
    "FileInput",
    "TEIDocument",
    "TEIPage",
    "TEIinput",
    "TextInput",
]



"""
The marimo notebook corresponding to the `remarx` application. The application
can be launched by running the command `remarx-app` or via marimo.

Example Usage:

    `remarx-app`

    `marimo run app.py`
"""

import marimo

__generated_with = "0.15.2"
app = marimo.App(width="medium")


@app.cell
def _():
    import csv
    import marimo as mo
    import pathlib
    import tempfile

    import logging
    import remarx
    from remarx.app.utils import create_header, create_temp_input, get_current_log_file
    from remarx.sentence.corpus.create import create_corpus
    from remarx.sentence.corpus import FileInput
    return (
        FileInput,
        create_corpus,
        create_header,
        create_temp_input,
        get_current_log_file,
        mo,
        pathlib,
        remarx,
        logging,
    )


@app.cell
def _(create_header):
    create_header()
    return


@app.cell
def _(get_current_log_file, logging):
    # Get log file path from already configured logging
    log_file_path = get_current_log_file()

    # Log that UI started
    logger = logging.getLogger("remarx-app")
    logger.info("Remarx Corpus Builder notebook started")

    return (log_file_path,)


@app.cell
def _(mo):
    mo.md(
        rf"""
    ## 📝 Sentence Corpus Builder
    Create a sentence corpus (`CSV`) from a text.
    This process can be run multiple times for different files (currently one file at a time).
    """
    )
    return


@app.cell
def _(FileInput, mo):
    mo.md(
        rf"""
    ### 1. Select Input Text

    Upload and select an input file (`{"`, `".join(FileInput.supported_types())}`) for sentence corpus creation.
    Currently, only a single file may be selected.
    """
    )
    return


@app.cell
def _(FileInput, mo):
    select_input = mo.ui.file(
        kind="area",
        filetypes=FileInput.supported_types(),
    )
    return (select_input,)


@app.cell
def _(mo, select_input):
    input_file = select_input.value[0] if select_input.value else None
    input_file_msg = f"`{input_file.name}`" if input_file else "None selected"
    input_callout_type = "success" if input_file else "warn"

    mo.callout(
        mo.vstack([select_input, mo.md(f"**Input File:** {input_file_msg}")]),
        kind=input_callout_type,
    )
    return (input_file,)


@app.cell
def _(mo):
    mo.md(
        r"""
    ### 2. Select Output Location

    Select the folder where the resulting sentence corpus file should be saved.
    The output CSV file will be named based on the input file.

    *To select a folder, click the file icon to the left of the folder's name.
    A checkmark will appear when a selection is made.
    Clicking anywhere else within the folder's row will cause the browser to navigate to this folder and subsequently display any folders *within* this folder.*
    """
    )
    return


@app.cell
def _(mo, pathlib):
    select_output_dir = mo.ui.file_browser(
        selection_mode="directory",
        multiple=False,
        initial_path=pathlib.Path.home(),
        filetypes=[],  # only show directories
    )
    return (select_output_dir,)


@app.cell
def _(mo, select_output_dir):
    output_dir = select_output_dir.value[0] if select_output_dir.value else None
    dir_callout_mode = "success" if output_dir else "warning"
    output_dir_msg = f"`{output_dir.path}`" if output_dir else "None selected"
    out_callout_type = "success" if output_dir else "warn"


    mo.callout(
        mo.vstack(
            [
                select_output_dir,
                mo.md(f"**Save Location:** {output_dir_msg}"),
            ],
        ),
        kind=out_callout_type,
    )
    return (output_dir,)


@app.cell
def _(mo):
    mo.md(
        r"""
    ### 3. Build Sentence Corpus

    Click the "Build Corpus" to run `remarx`.
    The sentence corpus for the input text will be saved as a CSV in the selected save location.
    This output file will have the same filename (but different file extension) as the selected input file.
    """
    )
    return


@app.cell
def _(input_file, mo, output_dir):
    # Determine inputs based on file & folder selections

    output_csv = (
        (output_dir.path / input_file.name).with_suffix(".csv")
        if input_file and output_dir
        else None
    )

    file_msg = (
        f"`{input_file.name}`" if input_file else "*Please select an input file*"
    )

    dir_msg = (
        f"`{output_dir.path}`"
        if output_dir
        else f"*Please select a save location*"
    )

    button = mo.ui.run_button(
        disabled=not (input_file and output_dir),
        label="Build Corpus",
        tooltip="Click to build sentence corpus",
    )

    mo.callout(
        mo.vstack(
            [
                mo.md(
                    f"""#### User Selections
                - **Input File:** {file_msg}
                - **Save Location**: {dir_msg}
            """
                ),
                button,
            ]
        ),
    )
    return button, output_csv


@app.cell
def _(button, create_corpus, create_temp_input, input_file, mo, output_csv):
    # Build Sentence Corpus
    building_msg = f'Click "Build Corpus" button to start'

    if button.value:
        spinner_msg = f"Building sentence corpus for {input_file.name}"
        with mo.status.spinner(title=spinner_msg) as _spinner:
            with create_temp_input(input_file) as temp_path:
                create_corpus(
                    temp_path, output_csv, filename_override=input_file.name
                )
        building_msg = f":white_check_mark: Sentence corpus saved to: {output_csv}"

    mo.md(building_msg).center()
    return


@app.cell
def _(mo, log_file_path):
    mo.md(f"Logs are being written to: {log_file_path}")
    return


if __name__ == "__main__":
    app.run()



import logging
import pathlib
import sys
from unittest.mock import Mock, patch

import pytest

from remarx.sentence.corpus.base_input import FileInput
from remarx.sentence.corpus.create import create_corpus, main


@patch("remarx.sentence.corpus.create.FileInput", spec=FileInput)
def test_create_corpus(mock_file_input, tmp_path: pathlib.Path):
    # Raise error when given dir
    with pytest.raises(ValueError, match=f"Input file {tmp_path} does not exist"):
        create_corpus(tmp_path, "out_path")

    # Raise error when given path to a non-existent file
    missing_file = tmp_path / "missing.txt"
    with pytest.raises(ValueError, match=f"Input file {missing_file} does not exist"):
        create_corpus(missing_file, "out_path")

    # Regular case
    ## File setup
    input_file = tmp_path / "input.txt"
    input_file.touch()
    out_csv = tmp_path / "out.csv"
    ## Mock input text
    mock_input = Mock()
    mock_file_input.create.return_value = mock_input
    mock_input.field_names = ["some", "field", "names"]
    mock_input.get_sentences.return_value = [
        {"some": "a", "field": "b", "names": "c"},
        {"some": "1", "field": "2", "names": "3"},
    ]

    create_corpus(input_file, out_csv)

    assert out_csv.is_file()
    mock_file_input.create.assert_called_once_with(input_file, filename_override=None)
    mock_input.get_sentences.assert_called_once_with()
    assert out_csv.read_text() == "some,field,names\na,b,c\n1,2,3\n"


@patch("remarx.sentence.corpus.create.FileInput", spec=FileInput)
def test_create_corpus_filename_override(mock_file_input, tmp_path: pathlib.Path):
    # test that filename override is passed through
    input_file = tmp_path / "tmp_foo_input_bar.txt"
    input_file.touch()
    out_csv = tmp_path / "out.csv"
    real_filename = "input.txt"
    create_corpus(input_file, out_csv, filename_override=real_filename)
    mock_file_input.create.assert_called_once_with(
        input_file, filename_override=real_filename
    )


@patch("remarx.sentence.corpus.create.configure_logging")
@patch("remarx.sentence.corpus.create.create_corpus", spec=create_corpus)
def test_main(mock_create_corpus, mock_config_logging):
    with patch("sys.argv", ["create_corpus.py", "input", "output"]):
        main()
        mock_create_corpus.assert_called_once_with(
            pathlib.Path("input"), pathlib.Path("output")
        )
        mock_config_logging.assert_called_once_with(sys.stdout, log_level=logging.INFO)

    mock_config_logging.reset_mock()
    # test verbose arg
    with patch("sys.argv", ["create_corpus.py", "input", "output", "-v"]):
        main()
        mock_config_logging.assert_called_once_with(sys.stdout, log_level=logging.DEBUG)



import pathlib
from collections.abc import Generator
from unittest.mock import Mock, patch

import pytest
from lxml.etree import Element

from remarx.sentence.corpus.base_input import FileInput, SectionType
from remarx.sentence.corpus.tei_input import TEI_TAG, TEIDocument, TEIinput, TEIPage

FIXTURE_DIR = pathlib.Path(__file__).parent / "fixtures"
TEST_TEI_FILE = FIXTURE_DIR / "sample_tei.xml"
TEST_TEI_WITH_FOOTNOTES_FILE = FIXTURE_DIR / "sample_tei_with_footnotes.xml"


def test_tei_tag():
    # test that tei tags object is constructed as expected
    assert TEI_TAG.pb == "{http://www.tei-c.org/ns/1.0}pb"


class TestTEIDocument:
    def test_init_from_file(self):
        tei_doc = TEIDocument.init_from_file(TEST_TEI_FILE)
        assert isinstance(tei_doc, TEIDocument)
        # fixture currently includes 4 pb tags, 2 of which are manuscript edition
        assert len(tei_doc.all_pages) == 4
        assert isinstance(tei_doc.all_pages[0], TEIPage)
        # first pb in sample is n=12
        assert tei_doc.all_pages[0].number == "12"

    def test_init_error(self, tmp_path: pathlib.Path):
        txtfile = tmp_path / "non-tei.txt"
        txtfile.write_text("this is not tei or xml")
        with pytest.raises(ValueError, match="Error parsing"):
            TEIDocument.init_from_file(txtfile)

    def test_pages(self):
        tei_doc = TEIDocument.init_from_file(TEST_TEI_FILE)
        # pages should be filtered to the standard edition only
        assert len(tei_doc.pages) == 2
        # for these pages, edition attribute is not present
        assert all(p.edition is None for p in tei_doc.pages)

    def test_pages_by_number(self):
        tei_doc = TEIDocument.init_from_file(TEST_TEI_FILE)
        assert isinstance(tei_doc.pages_by_number, dict)
        assert list(tei_doc.pages_by_number.values()) == tei_doc.pages
        assert list(tei_doc.pages_by_number.keys()) == [p.number for p in tei_doc.pages]
        first_page = tei_doc.pages[0]
        assert tei_doc.pages_by_number[first_page.number] == first_page


class TestTEIPage:
    def test_attributes(self):
        tei_doc = TEIDocument.init_from_file(TEST_TEI_FILE)
        # test first page and first manuscript page
        page = tei_doc.all_pages[0]
        ms_page = tei_doc.all_pages[1]

        assert page.number == "12"
        assert page.edition is None

        assert ms_page.number == "IX"
        assert ms_page.edition == "manuscript"

    @patch.object(TEIPage, "get_body_text")
    @patch.object(TEIPage, "get_footnote_text")
    def test_str(self, mock_get_footnote_text, mock_get_body_text):
        mock_get_body_text.return_value = "Mock body text"
        mock_get_footnote_text.return_value = "Mock footnote text"

        tei_doc = TEIDocument.init_from_file(TEST_TEI_FILE)
        page = tei_doc.all_pages[0]

        result = str(page)

        # Verify mocks were called (may be called multiple times)
        assert mock_get_body_text.called
        assert mock_get_footnote_text.called

        # should return both body text and footnote text, separated by double newlines
        assert result == "Mock body text\n\nMock footnote text"

    def test_get_body_text_no_footnotes(self):
        tei_doc = TEIDocument.init_from_file(TEST_TEI_FILE)
        # test first page
        page = tei_doc.all_pages[0]
        # includes some leading whitespace from <pb> and <p> tags
        # remove whitespace for testing for now
        text = page.get_body_text()

        # Should not contain leading or trailing whitespace
        assert text == text.strip()
        # first text content after the pb tag
        assert text.startswith("als in der ersten Darstellung.")  # codespell:ignore
        # last text content after the next standard pb tag
        assert text.endswith("entwickelten nur das Bild der eignen Zukunft!")
        # should not include editorial content
        assert "|" not in text
        assert "IX" not in text

    def test_get_body_text_with_footnotes(self):
        # test a sample page with footnotes to confirm footnote contents are excluded
        tei_doc = TEIDocument.init_from_file(TEST_TEI_WITH_FOOTNOTES_FILE)
        page_17 = next(p for p in tei_doc.all_pages if p.number == "17")

        body_text = page_17.get_body_text()
        assert body_text.startswith(
            "Der Reichthum der Gesellschaften"
        )  # codespell:ignore
        assert "1) Karl Marx:" not in body_text  # Footnote content should be excluded

    def test_get_body_text_line_numbers(self):
        tei_doc = TEIDocument.init_from_file(TEST_TEI_WITH_FOOTNOTES_FILE)
        page_17 = next(p for p in tei_doc.pages if p.number == "17")

        body_text = page_17.get_body_text()
        first_line_idx = body_text.index("Der Reichthum der Gesellschaften")
        second_line_idx = body_text.index("Die Waare ist zunächst")  # codespell:ignore

        assert page_17.get_body_text_line_number(first_line_idx) == 1
        assert page_17.get_body_text_line_number(second_line_idx) == 5
        assert page_17.get_body_text_line_number(second_line_idx + 10) == 5

    def test_get_body_text_line_number_without_cached_data(self):
        """Test get_body_text_line_number when line_number_by_offset doesn't exist yet."""
        tei_doc = TEIDocument.init_from_file(TEST_TEI_WITH_FOOTNOTES_FILE)
        page_17 = next(p for p in tei_doc.pages if p.number == "17")

        # Ensure the page doesn't have cached line number data
        assert not hasattr(page_17, "line_number_by_offset")

        # This should trigger the hasattr check and call get_body_text()
        line_number = page_17.get_body_text_line_number(0)

        # After the call, the attribute should be set
        assert hasattr(page_17, "line_number_by_offset")
        assert line_number == 1

    def test_get_body_text_multiple_lb_line_breaks(self):
        tei_doc = TEIDocument.init_from_file(TEST_TEI_WITH_FOOTNOTES_FILE)
        page_21 = next(p for p in tei_doc.pages if p.number == "21")

        body_text = page_21.get_body_text()

        assert "Fortgang der\nAccumulation" in body_text
        assert "derAccumulation" not in body_text

    def test_get_body_text_line_numbers_without_any_lb(self):
        tei_doc = TEIDocument.init_from_file(TEST_TEI_WITH_FOOTNOTES_FILE)
        page_22 = next(p for p in tei_doc.pages if p.number == "22")

        body_text = page_22.get_body_text()
        assert body_text.startswith("Eine Seite ohne markierte Zeilenumbrüche.")
        assert page_22.line_number_by_offset == {}
        assert page_22.get_body_text_line_number(0) is None

    def test_get_body_text_line_numbers_with_inline_markup(self):
        tei_doc = TEIDocument.init_from_file(TEST_TEI_WITH_FOOTNOTES_FILE)
        page_21 = next(p for p in tei_doc.pages if p.number == "21")

        body_text = page_21.get_body_text()

        inline_start = body_text.index("Schiedensten Proportionen")
        assert body_text[:inline_start].endswith("\n")
        assert page_21.get_body_text_line_number(inline_start) == 31

        assert "Dennoch bleibt sein\nTauschwerth" in body_text
        second_block_start = body_text.index("Dennoch bleibt sein\nTauschwerth")
        tauschwerth_idx = second_block_start + len("Dennoch bleibt sein\n")
        assert page_21.get_body_text_line_number(tauschwerth_idx) == 34
        unveraendert_idx = body_text.index("unverändert", tauschwerth_idx)
        assert page_21.get_body_text_line_number(unveraendert_idx) == 34

        # Ensure blank <lb/> entries still add newline placeholders and line mappings.
        assert "\n\nLeerzeile als separates Beispiel." in body_text  # codespell:ignore
        leerzeile_idx = body_text.index(
            "Leerzeile als separates Beispiel."  # codespell:ignore
        )  # codespell:ignore
        assert page_21.get_body_text_line_number(leerzeile_idx) == 39
        assert 38 in page_21.line_number_by_offset.values()

    def test_get_footnote_text_with_footnotes(self):
        tei_doc = TEIDocument.init_from_file(TEST_TEI_WITH_FOOTNOTES_FILE)
        page_17 = next(p for p in tei_doc.all_pages if p.number == "17")

        footnote_text = page_17.get_footnote_text()
        assert footnote_text.startswith("1) Karl Marx:")
        assert "Nicholas Barbon" in footnote_text
        assert (
            "Der Reichthum der Gesellschaften" not in footnote_text
        )  # Body text should be excluded

    def test_get_footnote_line_numbers(self):
        tei_doc = TEIDocument.init_from_file(TEST_TEI_WITH_FOOTNOTES_FILE)
        page_17 = next(p for p in tei_doc.pages if p.number == "17")

        footnotes = list(page_17.get_page_footnotes())
        # returns footnote xmlobject, which has a line number attribute
        assert footnotes[0].line_number == 17
        assert footnotes[1].line_number == 18

    def test_get_footnote_text_delimiter(self):
        # Test that footnotes are properly separated by double newlines
        tei_doc = TEIDocument.init_from_file(TEST_TEI_WITH_FOOTNOTES_FILE)
        page_17 = next(p for p in tei_doc.all_pages if p.number == "17")

        footnote_text = page_17.get_footnote_text()
        # Check that double newlines are present between footnotes
        # The fixture should have multiple footnotes to test this properly
        assert "\n\n" in footnote_text

    def test_is_footnote_content(self):
        # Test direct footnote elements
        footnote_ref = Element(TEI_TAG.ref, type="footnote")
        footnote_note = Element(TEI_TAG.note, type="footnote")
        regular_element = Element("p")

        assert TEIPage.is_footnote_content(footnote_ref)
        assert TEIPage.is_footnote_content(footnote_note)
        assert not TEIPage.is_footnote_content(regular_element)

        # Test nested elements within footnotes
        # Create sample XML tree to mimic the structure of a footnote:
        # <note type="footnote"><p><em>text</em></p></note>
        footnote_container = Element(
            TEI_TAG.note, type="footnote"
        )  # create a footnote container element
        paragraph = Element("p")  # create a paragraph element
        emphasis = Element("em")  # create an emphasis element

        footnote_container.append(
            paragraph
        )  # nest the paragraph element within the footnote container
        paragraph.append(
            emphasis
        )  # nest the emphasis element within the paragraph element

        # Test that nested elements are correctly identified as footnote content
        assert TEIPage.is_footnote_content(paragraph)
        assert TEIPage.is_footnote_content(emphasis)

        # Test element outside footnote structure
        standalone_p = Element("p")
        assert not TEIPage.is_footnote_content(standalone_p)

    def test_get_body_text_line_numbers_missing_lb_number(self):
        tei_doc = TEIDocument.init_from_file(TEST_TEI_WITH_FOOTNOTES_FILE)
        page_20 = next(p for p in tei_doc.pages if p.number == "20")

        body_text = page_20.get_body_text()
        assert "Erste Umschlagseite der Erstausgabe" in body_text, (
            "Fixture text missing expected caption content"
        )

        # No numbered line marker before the first characters, so None is expected
        assert page_20.get_body_text_line_number(0) is None

    def test_find_preceding_lb(self):
        tei_doc = TEIDocument.init_from_file(TEST_TEI_WITH_FOOTNOTES_FILE)
        # use xpath to get known elements from fixture doc for testing
        # all xpaths need tei namespace declared so we can use t: prefix
        tei_ns = {"namespaces": TEIDocument.ROOT_NAMESPACES}

        # lb 31 has an immediately preceding hi tag
        linebegin31 = tei_doc.node.xpath('.//t:lb[@n="31"]', **tei_ns)[0]
        inline_el = tei_doc.node.xpath(
            './/t:hi[preceding-sibling::t:lb[@n="31"]]', **tei_ns
        )[0]
        assert TEIPage.find_preceding_lb(inline_el) == linebegin31
        # test nested inline el
        linebegin2 = tei_doc.node.xpath('.//t:lb[@n="2"]', **tei_ns)[0]
        nested_inline_el = tei_doc.node.xpath(
            './/t:ref[@target="fn1-1"]/t:hi', **tei_ns
        )[0]
        assert TEIPage.find_preceding_lb(nested_inline_el) == linebegin2
        # no preceding lb should return none
        pagebegin2 = tei_doc.node.xpath('.//t:pb[@n="17"]', **tei_ns)[0]
        assert TEIPage.find_preceding_lb(pagebegin2) is None


class TestTEIinput:
    def test_init(self):
        tei_input = TEIinput(input_file=TEST_TEI_FILE)
        assert tei_input.input_file == TEST_TEI_FILE
        # xml is parsed as tei document
        assert isinstance(tei_input.xml_doc, TEIDocument)

    def test_field_names(self):
        # includes defaults from text input and adds page number, section type, and line number
        assert TEIinput.field_names == (
            *FileInput.field_names,
            "page_number",
            "section_type",
            "line_number",
        )

    def test_get_text(self):
        tei_input = TEIinput(input_file=TEST_TEI_FILE)
        text_result = tei_input.get_text()
        # should be a generator
        assert isinstance(text_result, Generator)
        text_result = list(text_result)
        # expect two pages
        assert len(text_result) == 2
        # result type is dictionary
        assert all(isinstance(txt, dict) for txt in text_result)
        # check for expected contents
        # - page text
        assert (
            text_result[0]["text"]
            .strip()
            .startswith("als in der ersten")  # codespell:ignore
        )
        assert text_result[1]["text"].strip().startswith("Aber abgesehn hiervon")
        # - page number
        assert text_result[0]["page_number"] == "12"
        assert text_result[1]["page_number"] == "13"
        # - section type
        assert text_result[0]["section_type"] == "text"
        assert text_result[1]["section_type"] == "text"

    def test_get_text_with_footnotes(self):
        tei_input = TEIinput(input_file=TEST_TEI_WITH_FOOTNOTES_FILE)
        text_chunks = list(tei_input.get_text())

        # Should get both text and footnote chunks for each page
        section_types = [chunk["section_type"] for chunk in text_chunks]
        assert "text" in section_types
        assert "footnote" in section_types

        # Check page numbers are set correctly
        assert all("page_number" in chunk for chunk in text_chunks)
        assert all(isinstance(chunk["text"], str) for chunk in text_chunks)

    def test_get_extra_metadata_line_numbers(self):
        tei_input = TEIinput(input_file=TEST_TEI_WITH_FOOTNOTES_FILE)
        text_chunks = list(tei_input.get_text())

        body_chunk = next(
            chunk
            for chunk in text_chunks
            if chunk["section_type"] == SectionType.TEXT.value
            and chunk["page_number"] == "17"
        )
        body_idx = body_chunk["text"].index(
            "Die Waare ist zunächst"  # codespell:ignore
        )  # codespell:ignore
        body_metadata = tei_input.get_extra_metadata(
            body_chunk, body_idx, body_chunk["text"]
        )
        assert body_metadata["line_number"] == 5

        footnote_chunk = next(
            chunk
            for chunk in text_chunks
            if chunk["section_type"] == SectionType.FOOTNOTE.value
            and chunk["page_number"] == "17"
            and chunk["text"].startswith("1) Karl Marx:")
        )
        footnote_metadata = tei_input.get_extra_metadata(
            footnote_chunk, 4, footnote_chunk["text"]
        )
        assert footnote_metadata["line_number"] == 17

    def test_get_extra_metadata_missing_page_returns_none(self):
        tei_input = TEIinput(input_file=TEST_TEI_FILE)
        metadata = tei_input.get_extra_metadata(
            {
                "page_number": "999",
                "section_type": SectionType.TEXT.value,
                "text": "",
            },
            0,
            "",
        )
        assert metadata["line_number"] is None

    @patch("remarx.sentence.corpus.base_input.segment_text")
    def test_get_sentences(self, mock_segment_text: Mock):
        tei_input = TEIinput(input_file=TEST_TEI_FILE)
        # segment text returns a tuple of character index, sentence text
        mock_segment_text.return_value = [(0, "Aber abgesehn hiervon")]
        sentences = tei_input.get_sentences()
        # expect a generator with one item, with the content added to the file
        assert isinstance(sentences, Generator)
        sentences = list(sentences)
        assert len(sentences) == 2  # 2 pages, one mock sentence each
        # method called once for each page of text
        assert mock_segment_text.call_count == 2
        assert all(isinstance(sentence, dict) for sentence in sentences)
        # file id set (handled by base input class)
        assert sentences[0]["file"] == TEST_TEI_FILE.name
        # page number set
        assert sentences[0]["page_number"] == "12"
        assert sentences[1]["page_number"] == "13"
        # sentence index is set and continues across pages
        assert sentences[0]["sent_index"] == 0
        assert sentences[1]["sent_index"] == 1

    @patch("remarx.sentence.corpus.base_input.segment_text")
    def test_get_sentences_with_footnotes(self, mock_segment_text: Mock):
        tei_input = TEIinput(input_file=TEST_TEI_WITH_FOOTNOTES_FILE)
        # segment text returns a tuple of character index, sentence text
        mock_segment_text.return_value = [(0, "Aber abgesehn hiervon")]
        sentences = tei_input.get_sentences()
        # expect a generator
        assert isinstance(sentences, Generator)
        sentences = list(sentences)
        # all should be dictionaries
        assert all(isinstance(sentence, dict) for sentence in sentences)
        # should have both text and footnote sections
        section_types = [s["section_type"] for s in sentences]
        assert "text" in section_types
        assert "footnote" in section_types



"""
Library for generating sentence embeddings from pretrained Sentence Transformer models.
"""

import logging
from timeit import default_timer as time

import numpy.typing as npt
from sentence_transformers import SentenceTransformer

logger = logging.getLogger(__name__)


def get_sentence_embeddings(
    sentences: list[str],
    model_name: str = "paraphrase-multilingual-mpnet-base-v2",
    show_progress_bar: bool = False,
) -> npt.NDArray:
    """
    Extract embeddings for each sentence using the specified pretrained Sentence
    Transformers model (default is paraphrase-multilingual-mpnet-base-v2).
    Returns a numpy array of the embeddings with shape [# sents, # dims].

    :param sentences: List of sentences to generate embeddings for
    :param model_name: Name of the pretrained sentence transformer model to use (default: paraphrase-multilingual-mpnet-base-v2)
    :return: 2-dimensional numpy array of normalized sentence embeddings with shape [# sents, # dims]
    """

    # Generate embeddings using the specified model
    start = time()
    model = SentenceTransformer(model_name)
    embeddings = model.encode(
        sentences,
        normalize_embeddings=True,
        show_progress_bar=show_progress_bar,
    )
    n_vecs = len(embeddings)
    elapsed_time = time() - start
    logger.info(f"Generated {n_vecs:,} embeddings in {elapsed_time:.1f} seconds")
    return embeddings



"""
Script for transforming an html page into a plaintext file.

NOTE: This script is tailored specifically for the Communist Manifesto webpages.

Example Usage:

    `uv run python parse_html.py in_html out_txt`
"""

import argparse
import pathlib
import re
import sys

from bs4 import BeautifulSoup, NavigableString, Tag


def get_tag_text(tag: Tag) -> str:
    """
    Generates the text for a given Beautiful Tag. Hyperlinks are ignored.

    :returns: str of the tag's generated text
    """
    out_text = ""
    for child_elt in tag.children:
        if isinstance(child_elt, NavigableString):
            out_text += child_elt
        else:
            # Skip hyperlink tags
            if child_elt.name != "a":
                out_text += child_elt.get_text()
    # Combine consecutive spaces & strip leading and trailing whitespace
    return re.sub(r"  *", " ", out_text).strip()


def get_html_text(in_html: pathlib.Path) -> str:
    """
    Extracts the text from the input html file. Footnotes are ignored.

    :returns: str of the extracted text for `in_html`
    """
    soup = BeautifulSoup(in_html.read_text(), "html.parser")
    body_soup = soup.body

    # Find starting paragraph text
    start_p = body_soup.find("p", class_="fst")

    # Start at previous section (h3) or title (h1) header
    start_elt = start_p.find_previous(re.compile(r"^h[1,3]"))
    text = get_tag_text(start_elt) + "\n\n"

    # Add each following element's text until the notes section is reached
    # Notes section is an h3 tag with text "Anmerkungen"
    for elt in start_elt.next_siblings:
        # Skip NavigableStrings
        if isinstance(elt, NavigableString):
            continue

        tag_name = elt.name
        tag_text = get_tag_text(elt).strip()

        # Stop when the notes section is reached
        if tag_name == "h3" and tag_text.startswith("Anmerkung"):
            break

        if tag_name == "p":
            # Case: Paragraphs
            if not tag_text or ("class" in elt and elt["class"] == "link"):
                # Skip tags without non-whitespace text or with links
                continue
            text += f"{tag_text}\n\n"
        elif re.fullmatch(r"h\d", tag_name):
            # Case: Headers
            text += f"{tag_text}\n\n"
        elif tag_name == "ol":
            # Case: Ordered lists
            for i, list_elt in enumerate(elt.find_all("li", recursive=False)):
                text += f"{i + 1}. {get_tag_text(list_elt)}\n"
            text += "\n"
        else:
            print(f"Warning: Skipping unsupported tag {tag_name}")
    return text.strip()


def write_html_text(in_html: pathlib.Path, out_txt: pathlib.Path) -> None:
    """
    Extracts the text from the input html file and saves the resulting file at
    the specified filepath. Calls :meth:`get_html_text`.

    :raises ValueError: if `in_html` does not exist
    :raises ValueError: if `out_txt` already exists
    """
    # Validate input args
    if not in_html.is_file():
        raise ValueError("Input file does not exist")
    if out_txt.is_file():
        raise ValueError("Output file exists. Not overwriting.")

    out_txt.write_text(get_html_text(in_html))


def main() -> None:
    """
    Command-line access for extracting the text from an html page
    """
    parser = argparse.ArgumentParser(
        description="Extract the text from a given html page and save as a plaintext file",
    )
    parser.add_argument(
        "in_file",
        type=pathlib.Path,
        help="Path to input html file",
    )
    parser.add_argument(
        "out_txt",
        type=pathlib.Path,
        help="Path to output plaintext file",
    )

    args = parser.parse_args()
    try:
        write_html_text(args.in_file, args.out_txt)
    except ValueError as e:
        print(f"ERROR: {e}", file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()



"""
Provides functionality to break down input text into individual
sentences and return them as tuples containing the character index where each
sentence begins and the sentence text itself.
"""

import stanza
from stanza import DownloadMethod


def segment_text(text: str, language: str = "de") -> list[tuple[int, str]]:
    """
    Segment a string of text into sentences with character indices.

    :param text: Input text to be segmented into sentences
    :param language: Language code for the Stanza pipeline
    :return: List of tuples where each tuple contains (start_char_index, sentence_text)
    """
    # Initialize the NLP pipeline for sentence segmentation
    # Use minimal processors (tokenize) for sentence segmentation only
    segmenter = stanza.Pipeline(
        lang=language,
        processors="tokenize",
        download_method=DownloadMethod.REUSE_RESOURCES,
    )

    processed_doc = segmenter(text)

    # Extract sentences with character-level indices
    return [
        (sentence.tokens[0].start_char, sentence.text)
        for sentence in processed_doc.sentences
    ]



import csv
import logging
import re
from unittest.mock import Mock, call, patch

import numpy as np
import polars as pl
import pytest
from polars.testing import assert_frame_equal
from voyager import Index, Space

from remarx.quotation.pairs import (
    build_vector_index,
    compile_quote_pairs,
    find_quote_pairs,
    get_sentence_pairs,
    load_sent_df,
)


@patch("remarx.quotation.pairs.Index")
def test_build_vector_index(mock_index_class, caplog):
    mock_index = Mock(spec=Index)
    mock_index_class.return_value = mock_index
    test_embeddings = np.ones([10, 50])
    mock_index.num_elements = 10

    # Default case
    result = build_vector_index(test_embeddings)
    assert result is mock_index
    mock_index_class.assert_called_once_with(
        Space.InnerProduct, num_dimensions=50, max_elements=10
    )
    assert mock_index.add_items.call_count == 1

    # can't use assert call due to numpy array equality check
    # get args and check for expected match
    args, _kwargs = mock_index.add_items.call_args
    assert np.array_equal(args[0], test_embeddings)

    # Check logging
    caplog.clear()
    with caplog.at_level(logging.INFO):
        _ = build_vector_index(test_embeddings)
    assert len(caplog.record_tuples) == 1
    assert caplog.record_tuples[0][1] == logging.INFO
    expected_msg = r"Created index with 10 items and 50 dimensions in \d+.\d seconds"
    assert re.fullmatch(expected_msg, caplog.record_tuples[0][2])


@patch("remarx.quotation.pairs.get_sentence_embeddings")
@patch("remarx.quotation.pairs.build_vector_index")
def test_get_sentence_pairs(mock_build_index, mock_embeddings, caplog):
    # setup mock index
    mock_index = Mock(spec=Index)
    # list of lists of ids, distances
    test_results = ([[0], [5], [1]], [[0.7], [0.4], [0.18]])
    mock_index.query.return_value = test_results
    mock_build_index.return_value = mock_index

    # setup mock embeddings
    original_vecs = np.array([[5], [10]])
    reuse_vecs = np.array([[0], [1], [2]])
    mock_embeddings.side_effect = [original_vecs, reuse_vecs]

    # Case: Basic
    expected = pl.DataFrame(
        [{"reuse_index": 2, "original_index": 1, "match_score": 0.18}]
    ).cast({"reuse_index": pl.UInt32})  # cast to match row index type
    results = get_sentence_pairs("original_sents", "reuse_sents", 0.2)
    assert_frame_equal(results, expected)
    ## check mock calls
    assert mock_embeddings.call_count == 2
    mock_embeddings.assert_has_calls(
        [
            call("original_sents", show_progress_bar=False),
            call("reuse_sents", show_progress_bar=False),
        ]
    )
    mock_build_index.assert_called_once_with(original_vecs)
    assert mock_index.query.call_count == 1
    mock_index.query.assert_called_with(reuse_vecs, k=1)

    # Case: Check logging
    mock_embeddings.side_effect = [original_vecs, reuse_vecs]
    caplog.clear()
    with caplog.at_level(logging.INFO):
        _ = get_sentence_pairs("original_sents", "reuse_sents", 0.2)

    # currently all logging is info level
    assert len(caplog.record_tuples) == 4
    assert all(log[1] == logging.INFO for log in caplog.record_tuples)
    # check log messages for expected text;
    # order agnostic; just check for presence of expected messages
    log_messages = [log[2] for log in caplog.record_tuples]
    assert "Now generating sentence embeddings" in log_messages
    assert any(
        re.fullmatch(r"Generated 5 sentence embeddings in \d+\.\d seconds", log)
        for log in log_messages
    )
    assert any(
        re.fullmatch(r"Queried 3 sentence embeddings in \d+\.\d seconds", log)
        for log in log_messages
    )
    assert "Identified 1 sentence pair under score cutuff 0.2" in log_messages


def test_load_sent_df(tmp_path):
    # Setup test sentence corpus
    test_csv = tmp_path / "sent_corpus.csv"
    test_data = {
        "sent_id": ["a", "b", "c"],
        "text": ["foo", "bar", "baz"],
    }

    test_df = pl.DataFrame(test_data)
    test_df.write_csv(test_csv)

    # Basic case (minimal fields)
    ## No prefix
    expected_data = {
        "index": [0, 1, 2],
        "id": test_data["sent_id"],
        "text": test_data["text"],
    }
    expected = pl.DataFrame(expected_data)
    result = load_sent_df(test_csv)
    assert_frame_equal(result, expected, check_dtypes=False)
    ## With prefix
    pfx_expected = pl.DataFrame({f"test_{k}": v for k, v in expected_data.items()})
    result = load_sent_df(test_csv, "test_")
    assert_frame_equal(result, pfx_expected, check_dtypes=False)

    # Case additional metadata fields
    test_df = test_df.with_columns(
        pl.Series("other", ["x", "y", "z"]),
        pl.Series("misc", [0, 1, 2]),
    )
    test_df.write_csv(test_csv)
    ## No prefix
    expected = expected.with_columns(
        pl.Series("other", ["x", "y", "z"]),
        # Values are strings because no schema inference
        pl.Series("misc", ["0", "1", "2"]),
    )
    result = load_sent_df(test_csv)
    assert_frame_equal(result, expected, check_dtypes=False)
    ## With prefix
    pfx_expected = pfx_expected.with_columns(
        expected.get_column("other").rename("test_other"),
        expected.get_column("misc").rename("test_misc"),
    )
    result = load_sent_df(test_csv, "test_")
    assert_frame_equal(result, pfx_expected, check_dtypes=False)


def test_compile_quote_pairs():
    # Both corpora include unmatched sentences to ensure that the output
    # does not include unmatched sentences
    reuse_df = pl.DataFrame(
        # a, c are unmatched
        {
            "reuse_index": [0, 1, 2, 3, 4],
            "reuse_id": ["a", "b", "c", "d", "e"],
            "reuse_text": ["0", "1", "2", "3", "4"],
            "reuse_other": [4, 3, 2, 1, 0],
        }
    )
    orig_df = pl.DataFrame(
        # B is unmatched
        {
            "original_index": [0, 1, 2],
            "original_id": ["A", "B", "C"],
            "original_text": ["0", "1", "2"],
            "original_other": [2, 1, 0],
        }
    )

    # Includes two pairs with the same original sentence (A)
    detected_pairs = pl.DataFrame(
        {
            "reuse_index": [1, 3, 4],
            "original_index": [2, 0, 0],
            "match_score": [0.1, 0.225, 0.01],
        }
    )

    # Expecting 3 quote pairs: b-C, d-A, e-A
    expected = pl.DataFrame(
        {
            "match_score": [0.1, 0.225, 0.01],
            "reuse_id": ["b", "d", "e"],
            "reuse_text": ["1", "3", "4"],
            "reuse_other": [3, 1, 0],
            "original_id": ["C", "A", "A"],
            "original_text": ["2", "0", "0"],
            "original_other": [0, 2, 2],
        }
    )

    result = compile_quote_pairs(orig_df, reuse_df, detected_pairs)
    print(f"result: {result.columns}")
    print(f"expected: {expected.columns}")
    assert_frame_equal(result, expected, check_row_order=False)


@patch("remarx.quotation.pairs.compile_quote_pairs")
@patch("remarx.quotation.pairs.get_sentence_pairs")
@patch("remarx.quotation.pairs.load_sent_df")
def test_find_quote_pairs(
    mock_load_df, mock_sent_pairs, mock_compile_pairs, caplog, tmp_path
):
    # setup mocks
    orig_texts = ["some", "text"]
    reuse_texts = ["some", "other", "texts"]
    orig_df = pl.DataFrame({"original_text": orig_texts})
    reuse_df = pl.DataFrame({"reuse_text": reuse_texts})
    mock_load_df.side_effect = [orig_df, reuse_df]
    mock_sent_pairs.return_value = ["sent_pairs"]
    mock_compile_pairs.return_value = pl.DataFrame({"foo": 1, "bar": "a"})

    # Basic
    out_csv = tmp_path / "out.csv"
    find_quote_pairs("original", "reuse", out_csv)
    assert out_csv.read_text() == "foo,bar\n1,a\n"
    ## check mocks
    assert mock_load_df.call_count == 2
    mock_sent_pairs.assert_called_once_with(
        orig_texts, reuse_texts, 0.225, show_progress_bar=False
    )
    mock_compile_pairs.assert_called_once_with(orig_df, reuse_df, ["sent_pairs"])

    ## check logging
    mock_load_df.side_effect = [orig_df, reuse_df]
    with caplog.at_level(logging.INFO):
        caplog.clear()
        find_quote_pairs("original", "reuse", out_csv)
    logs = caplog.record_tuples
    assert len(logs) == 1
    assert logs[0][0] == "remarx.quotation.pairs"
    assert logs[0][1] == logging.INFO
    ### check log messages
    assert re.fullmatch(f"Saved 1 quote pairs to {out_csv}", logs[0][2])

    # Specify cutoff
    mock_load_df.side_effect = [orig_df, reuse_df]
    mock_sent_pairs.reset_mock()
    out_csv = tmp_path / "cutoff.csv"
    find_quote_pairs("original", "reuse", out_csv, score_cutoff=0.4)
    mock_sent_pairs.assert_called_once_with(
        orig_texts, reuse_texts, 0.4, show_progress_bar=False
    )

    # Case: show progress bar
    mock_load_df.side_effect = [orig_df, reuse_df]
    mock_sent_pairs.reset_mock()
    out_csv = tmp_path / "progress.csv"
    find_quote_pairs("original", "reuse", out_csv, show_progress_bar=True)
    # check mocks
    mock_sent_pairs.assert_called_once_with(
        orig_texts, reuse_texts, 0.225, show_progress_bar=True
    )

    # Case no results
    mock_load_df.side_effect = [orig_df, reuse_df]
    mock_sent_pairs.return_value = []
    with caplog.at_level(logging.INFO):
        caplog.clear()
        find_quote_pairs("original", "reuse", out_csv)
    assert len(caplog.record_tuples) == 1
    log = caplog.record_tuples[0]
    assert log[1] == logging.INFO
    assert "No sentence pairs for score cutoff = 0.225" in log[2]


def test_find_quote_pairs_integration(tmp_path):
    """
    Tests the full quote detection pipeline. Checks that all functions within this
    library work as expected in combination. This tests behavior that is otherwise
    masked by mocking.
    """
    test_orig = pl.DataFrame(
        [
            {
                "sent_id": "B",
                "corpus": "original",
                "text": "Und nun sollen seine Geister Auch nach meinem Willen leben.",
            },
            {
                "sent_id": "A",
                "corpus": "original",
                "text": "Hat der alte Hexenmeister Sich doch einmal wegbegeben!",
            },
            {
                "sent_id": "C",
                "corpus": "original",
                "text": "Seine Wort und Werke Merkt ich und den Brauch, Und mit Geistesstärke Tu ich Wunder auch.",
            },
        ]
    )
    test_reuse = pl.DataFrame(
        [
            {
                "sent_id": "a",
                "corpus": "reuse",
                "text": "Hat der alte Hexenmeister Sich doch einmal wegbegeben!",
            },
            {"sent_id": "b", "text": "Komm zurück zu mir", "corpus": "reuse"},
        ]
    )
    # Create files
    orig_csv = tmp_path / "original.csv"
    test_orig.write_csv(orig_csv)
    reuse_csv = tmp_path / "reuse.csv"
    test_reuse.write_csv(reuse_csv)

    out_csv = tmp_path / "out.csv"
    find_quote_pairs(orig_csv, reuse_csv, out_csv)
    with out_csv.open(newline="") as file:
        reader = csv.DictReader(file)
        results = list(reader)
        assert len(results) == 1
        print(results[0].keys())
        assert list(results[0].keys()) == [
            "match_score",
            "reuse_id",
            "reuse_text",
            "reuse_corpus",
            "original_id",
            "original_text",
            "original_corpus",
        ]
        assert results[0]["reuse_id"] == "a"
        assert results[0]["original_id"] == "A"
        # Need to specify the relative tolerance because 0 is a special case
        assert float(results[0]["match_score"]) == pytest.approx(0, rel=1e-6, abs=1e-6)



"""
Functionality related to parsing MEGA TEI/XML content with the
goal of creating a sentence corpora with associated metadata
from the TEI.
"""

import logging
import pathlib
import re
from collections import namedtuple
from collections.abc import Generator
from dataclasses import dataclass, field
from functools import cached_property
from timeit import default_timer as time
from typing import Any, ClassVar, NamedTuple, Self

from lxml.etree import XMLSyntaxError, _Element
from neuxml import xmlmap

from remarx.sentence.corpus.base_input import FileInput, SectionType

logger = logging.getLogger(__name__)


TEI_NAMESPACE = "http://www.tei-c.org/ns/1.0"

# namespaced tags look like {http://www.tei-c.org/ns/1.0}tagname
# create a named tuple of short tag name -> namespaced tag name
TagNames: NamedTuple = namedtuple(
    "TagNames", ("pb", "lb", "note", "add", "label", "ref", "div3", "text", "p")
)
TEI_TAG = TagNames(**{tag: f"{{{TEI_NAMESPACE}}}{tag}" for tag in TagNames._fields})
"Convenience access to namespaced TEI tag names"


class BaseTEIXmlObject(xmlmap.XmlObject):
    """
    Base class for TEI XML objects with TEI namespace included in root namespaces,
    for use in XPath expressions.
    """

    ROOT_NAMESPACES: ClassVar[dict[str, str]] = {"t": TEI_NAMESPACE}


class TEIFootnote(BaseTEIXmlObject):
    """XmlObject class for footnotes."""

    line_number = xmlmap.IntegerField("./t:lb[1]/@n")
    "Line number where this footnote begins, based on first TEI line beginning (`lb`) within this note"

    # use xmlmap StringField method with normalize=True to collapse whitespace in the footnote text
    text = xmlmap.StringField(".", normalize=True)
    "Normalized text content for the footnote (collapses whitespace)"


class TEIPage(BaseTEIXmlObject):
    """
    Custom :class:`neuxml.xmlmap.XmlObject` instance for a page
    of content within a TEI XML document.
    """

    number = xmlmap.StringField("@n")
    "page number"
    edition = xmlmap.StringField("@ed")
    "page edition, if any"

    # page beginning tags delimit content instead of containing it;
    # use following axis to find all text nodes following this page beginning
    text_nodes = xmlmap.StringListField("following::text()")
    "list of all text nodes following this tag"

    # fetch footnotes after the current page break; will filter them in Python later
    # pb is a delimiter (not a container), so "following::note" returns all later footnotes
    following_footnotes = xmlmap.NodeListField(
        "following::t:note[@type='footnote']", TEIFootnote
    )
    "list of footnote elements within this page and following pages"

    next_page = xmlmap.NodeField(
        "following::t:pb[not(@ed)][1]",
        "self",
    )
    "the next standard page break after this one, or None if this is the last page"

    @staticmethod
    def is_footnote_content(el: _Element) -> bool:
        """
        Helper function that checks if an element or any of its ancestors is footnote content.
        """
        if (
            el.tag in [TEI_TAG.ref, TEI_TAG.note]
            and el.attrib.get("type") == "footnote"
        ):
            return True
        return any(
            TEIPage.is_footnote_content(ancestor) for ancestor in el.iterancestors()
        )

    def get_page_footnotes(self) -> list[TEIFootnote]:
        """
        Filters footnotes to keep only the footnotes that belong to this page.
        Only includes footnotes that occur between this pb and the next standard pb[not(@ed)].
        """
        page_footnotes: list[TEIFootnote] = []

        for footnote in self.following_footnotes:
            # If we have a next page and this footnote belongs to it, we're done
            if self.next_page and footnote in self.next_page.following_footnotes:
                break
            page_footnotes.append(footnote)

        return page_footnotes

    def get_body_text_line_number(self, char_pos: int) -> int | None:
        """
        Return the TEI line number for the line at or before `char_pos`.
        Returns None if no line number can be determined.
        """
        if not hasattr(self, "line_number_by_offset"):
            self.get_body_text()

        # When there are no line breaks with line numbers, return None
        if not self.line_number_by_offset:
            return None

        line_number = None
        for offset, ln in self.line_number_by_offset.items():
            if offset > char_pos:
                break
            line_number = ln
        return line_number

    @staticmethod
    def find_preceding_lb(element: _Element) -> _Element | None:
        """
        Find the closest preceding <lb> element for an element.
        Needed to find the <lb/> relative to immediately following
        inline markup, e.g. <lb n="31"/><hi>text ...</hi>
        """

        # First, iterate over preceding siblings;
        # Limit to TEI nodes to avoid iterating over non-tag nodes like comments
        for sibling in element.itersiblings(f"{{{TEI_NAMESPACE}}}*", preceding=True):
            if sibling.tag == TEI_TAG.lb:
                return sibling
            # if we hit a preceding paragraph, stop iterating (beyond inline text)
            if sibling.tag == TEI_TAG.p:
                break

        # if not found, try on the parent element
        parent = element.getparent()
        # if no parent, or hit a text element, we've gone too far; bail out
        if parent is None or parent.tag == TEI_TAG.text:
            return None
        return TEIPage.find_preceding_lb(element.getparent())

    def get_body_text(self) -> str:
        """
        Extract body text content for this page, excluding footnotes and editorial content.
        While collecting the text, build a mapping of character offsets to TEI line numbers.
        """
        body_text_parts: list[str] = []
        self.line_number_by_offset: dict[int, int] = {}
        char_offset = 0

        last_lb_el = None

        for text in self.text_nodes:
            # text here is an lxml smart string, which preserves context
            # in the xml tree and is associated with a parent tag.
            parent = text.getparent()

            # stop iterating when we hit the next page break;
            if self.next_page and parent == self.next_page.node:
                break

            # Skip this text node if it's inside a footnote tag
            if self.is_footnote_content(parent):
                continue

            # omit editorial content (e.g. original page numbers)
            if (
                parent.tag == TEI_TAG.add
                or (parent.tag == TEI_TAG.label and parent.get("type") == "mpb")
            ) and (text.is_text or (text.is_tail and text.strip() == "")):
                # omit if text is inside an editorial tag (is_text)
                # OR if text comes immediately after (is_tail) and is whitespace only
                continue

            cleaned_text = re.sub(r"\s*\n\s*", "\n", str(text))

            # Use lstrip() to strip leading whitespace from the very first text fragment
            # before concatenation to avoid counting leading newlines toward `char_offset`,
            # skewing line lookups.
            if not body_text_parts:
                cleaned_text = cleaned_text.lstrip()

            # check for line begin tag; could be direct parent
            # but in cases where <lb> is immediately followed by inline markup,
            # it may be skipped due to having no tail text
            preceding_lb = None
            if parent.tag == TEI_TAG.lb:
                preceding_lb = parent
            else:
                preceding_lb = self.find_preceding_lb(parent)

            if preceding_lb is not None and preceding_lb is not last_lb_el:
                # store the line number and character offset
                line_number = preceding_lb.get("n")
                self.line_number_by_offset[char_offset] = (
                    int(line_number) if line_number else None
                )
                # ensure text separated by <lb\> has a newline
                # if there is a preceding text segment and it does not end
                # with a newline, add one to the current text
                if body_text_parts and not body_text_parts[-1].endswith("\n"):
                    cleaned_text = f"\n{cleaned_text}"

                # set this element as the last lb handled, so we don't duplicate
                last_lb_el = preceding_lb

            if not cleaned_text:
                continue

            body_text_parts.append(cleaned_text)
            char_offset += len(cleaned_text)

        # join text parts and trim trailing whitespace
        body_text = "".join(body_text_parts).rstrip()

        return body_text

    def get_footnote_text(self) -> str:
        """
        Get all footnote content as a single string, with footnotes separated by double newlines.
        """
        return "\n\n".join(fn.text for fn in self.get_page_footnotes())

    def __str__(self) -> str:
        """
        Page text contents as a string, with body text and footnotes.
        """
        return f"{self.get_body_text()}\n\n{self.get_footnote_text()}"


class TEIDocument(BaseTEIXmlObject):
    """
    Custom :class:`neuxml.xmlmap.XmlObject` instance for TEI XML document.
    Customized for MEGA TEI XML.
    """

    all_pages = xmlmap.NodeListField("//t:text//t:pb", TEIPage)
    """List of page objects, identified by page begin tag (pb). Includes all
    pages (standard and manuscript edition), because the XPath is significantly
    faster without filtering."""

    @cached_property
    def pages(self) -> list[TEIPage]:
        """
        Standard pages for this document.  Returns a list of TEIPage objects
        for this document, omitting any pages marked as manuscript edition.
        """
        # it's more efficient to filter in python than in xpath
        return [page for page in self.all_pages if page.edition != "manuscript"]

    @cached_property
    def pages_by_number(self) -> dict[str, TEIPage]:
        """Dictionary lookup of standard pages by page number."""
        return {page.number: page for page in self.pages}

    @classmethod
    def init_from_file(cls, path: pathlib.Path) -> Self:
        """
        Class method to initialize a new :class:`TEIDocument` from a file.
        """
        try:
            return xmlmap.load_xmlobject_from_file(path, cls)
        except XMLSyntaxError as err:
            raise ValueError(f"Error parsing {path} as XML") from err


@dataclass
class TEIinput(FileInput):
    """
    Input class for TEI/XML content.  Takes a single input file,
    and yields text content by page, with page number.
    Customized for MEGA TEI/XML: follows standard edition page numbering
    and ignores pages marked as manuscript edition.
    """

    xml_doc: TEIDocument = field(init=False)
    "Parsed XML document; initialized from inherited input_file"

    field_names: ClassVar[tuple[str, ...]] = (
        *FileInput.field_names,
        "page_number",
        "section_type",
        "line_number",
    )
    "List of field names for sentences from TEI XML input files"

    file_type = ".xml"
    "Supported file extension for TEI/XML input"

    def __post_init__(self) -> None:
        """
        After default initialization, parse the input file as a
         [TEIDocument][remarx.sentence.corpus.tei_input.TEIDocument]
        and store it as [xml_doc][remarx.sentence.corpus.tei_input.TEIinput.xml_doc].
        """
        # parse the input file as xml and save the result
        self.xml_doc = TEIDocument.init_from_file(self.input_file)

    def get_text(self) -> Generator[dict[str, str]]:
        """
        Get document content as plain text. The document's content is yielded in segments
        with each segment corresponding to a dictionary of containing its text content,
        page number and section type ("text" or "footnote").
        Body text is yielded once per page, while each footnote is yielded individually.

        :returns: Generator with dictionaries of text content, with page number and section_type ("text" or "footnote").
        """
        # yield body text and footnotes content chunked by page with page number
        start = time()
        for page in self.xml_doc.pages_by_number.values():
            page_start = time()

            body_text = page.get_body_text()
            if body_text:
                yield {
                    "text": body_text,
                    "page_number": page.number,
                    "section_type": SectionType.TEXT.value,
                }

            # Yield each footnote individually to enforce separate sentence segmentation
            # so that separate footnotes cannot be combined into a single sentence
            for footnote in page.get_page_footnotes():
                yield {
                    "text": footnote.text,
                    "page_number": page.number,
                    "section_type": SectionType.FOOTNOTE.value,
                    "line_number": footnote.line_number,
                }

            page_elapsed_time = time() - page_start
            logger.debug(
                f"Processing page {page.number} in {page_elapsed_time:.2f} seconds"
            )

        elapsed_time = time() - start
        logger.info(
            f"Processed {self.file_name} with {len(self.xml_doc.pages)} pages in {elapsed_time:.1f} seconds"
        )

    def get_extra_metadata(
        self, chunk_info: dict[str, Any], char_idx: int, sentence: str
    ) -> dict[str, Any]:
        """
        Calculate extra metadata including line number for a sentence in TEI documents
        based on the character position within the text chunk (page body or footnote).

        :returns: Dictionary with line_number for the sentence (None if not found)
        """
        # If line_number is already in chunk_info (e.g., for footnotes), use it directly
        if "line_number" in chunk_info:
            return {"line_number": chunk_info["line_number"]}

        # Otherwise, calculate it for body text based on character position
        page_number = chunk_info["page_number"]
        page = self.xml_doc.pages_by_number.get(page_number)

        if page:
            line_number = page.get_body_text_line_number(char_idx)
            return {"line_number": line_number}

        return {"line_number": None}



"""
Utility functions for the remarx package
"""

import io
import logging
import pathlib
from datetime import datetime
from typing import TextIO


def configure_logging(
    log_destination: pathlib.Path | TextIO | None = None,
    log_level: int = logging.INFO,
    stanza_log_level: int = logging.ERROR,
) -> pathlib.Path | None:
    """
    Configure logging for the remarx application.
    Supports logging to any text stream, a specified file, or auto-generated timestamped file.

    :param log_destination: Where to write logs. Can be:
        - None (default): Creates a timestamped log file in ./logs/ directory
        - pathlib.Path: Write to the specified file path
        - Any io.TextIOBase (e.g., sys.stdout, sys.stderr, or any io.TextIOBase): Write to the given stream
    :param log_level: Logging level for remarx logger (default to logging.INFO)
    :param stanza_log_level: Logging level for stanza logger (default to logging.ERROR)
    :return: Path to the created log file if file logging is used, None if stream logging
    """

    log_file_path: pathlib.Path | None = None
    config_output_opts: dict
    if log_destination is None:
        # Default: create timestamped log file under cwd / logs/
        log_dir = pathlib.Path.cwd() / "logs"
        log_dir.mkdir(exist_ok=True)
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        log_file_path = log_dir / f"remarx_{timestamp}.log"
        config_output_opts = {"filename": log_file_path, "encoding": "utf-8"}
    elif isinstance(log_destination, io.TextIOBase):
        # Only allow io.TextIOBase instances as streams (includes sys.stdout, sys.stderr)
        config_output_opts = {"stream": log_destination}
    else:
        # File logging to specified path
        log_file_path = log_destination
        log_file_path.parent.mkdir(parents=True, exist_ok=True)
        config_output_opts = {"filename": log_file_path, "encoding": "utf-8"}

    # Use the lowest of the requested levels so debug logs are captured when any
    # component (e.g., stanza) is set to DEBUG
    effective_level = min(log_level, stanza_log_level)

    logging.basicConfig(
        level=effective_level,
        format="[%(asctime)s] %(levelname)s:%(name)s::%(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        force=True,
        **config_output_opts,
    )

    # Configure stanza logging level
    logging.getLogger("stanza").setLevel(stanza_log_level)

    return log_file_path
